"""Utility to hold the file-level pipeline execution plan."""

from __future__ import annotations

import pickle
import re
from dataclasses import dataclass, field

import hats.pixel_math.healpix_shim as hp
import numpy as np
from hats import pixel_math
from hats.io import file_io
from hats.pixel_math.healpix_pixel import HealpixPixel
from hats.pixel_math.sparse_histogram import HistogramAggregator, SparseHistogram
from numpy import frombuffer
from upath import UPath

import hats_import.file_io as import_io
from hats_import.pipeline_resume_plan import PipelineResumePlan


@dataclass
class ResumePlan(PipelineResumePlan):
    """Container class for holding the state of each file in the pipeline plan."""

    input_paths: list[UPath] = field(default_factory=list)
    """Resolved list of all files that will be used in the importer"""
    map_files: list[tuple[str, str]] = field(default_factory=list)
    """List of files (and job keys) that have yet to be mapped"""
    split_keys: list[tuple[str, str]] = field(default_factory=list)
    """Set of files (and job keys) that have yet to be split"""
    destination_pixel_map: dict[HealpixPixel, int] | None = None
    """Destination pixels and their expected final count"""
    threshold_mode: str = "row_count"
    """Which mode to use for partitioning: 'row_count' or 'mem_size'.
    Determines whether to create additional mem_size histogram."""
    should_run_mapping: bool = True
    should_run_splitting: bool = True
    should_run_reducing: bool = True
    should_run_finishing: bool = True

    MAPPING_STAGE = "mapping"
    SPLITTING_STAGE = "splitting"
    REDUCING_STAGE = "reducing"

    ROW_COUNT_HISTOGRAM_BINARY_FILE = "row_count_mapping_histogram.npz"
    ROW_COUNT_HISTOGRAMS_DIR = "row_count_histograms"
    MEM_SIZE_HISTOGRAM_BINARY_FILE = "mem_size_mapping_histogram.npz"
    MEM_SIZE_HISTOGRAMS_DIR = "mem_size_histograms"

    ALIGNMENT_FILE = "alignment.pickle"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        resume: bool = True,
        progress_bar: bool = True,
        simple_progress_bar: bool = False,
        input_paths=None,
        tmp_path=None,
        tmp_base_path: UPath | None = None,
        delete_resume_log_files: bool = True,
        delete_intermediate_parquet_files: bool = True,
        run_stages: list[str] | None = None,
        import_args=None,
    ):
        if import_args:
            super().__init__(pipeline_name="catalog", **import_args.resume_kwargs_dict())
            if import_args.debug_stats_only:
                run_stages = ["mapping", "finishing"]
            self.input_paths = import_args.input_paths

            # Set threshold_mode based on byte_pixel_threshold
            if hasattr(import_args, "byte_pixel_threshold") and import_args.byte_pixel_threshold is not None:
                self.threshold_mode = "mem_size"
        else:
            super().__init__(
                resume=resume,
                progress_bar=progress_bar,
                simple_progress_bar=simple_progress_bar,
                tmp_path=tmp_path,
                tmp_base_path=tmp_base_path,
                delete_resume_log_files=delete_resume_log_files,
                delete_intermediate_parquet_files=delete_intermediate_parquet_files,
            )
            self.input_paths = input_paths
        self.gather_plan(run_stages)

    def gather_plan(self, run_stages: list[str] | None = None):
        """Initialize the plan."""
        with self.print_progress(total=4, stage_name="Planning") as step_progress:
            ## Make sure it's safe to use existing resume state.
            super().safe_to_resume()
            step_progress.update(1)

            ## Validate existing resume state.
            ## - if a later stage is complete, the earlier stages should be complete too.
            mapping_done = self.done_file_exists(self.MAPPING_STAGE)
            splitting_done = self.done_file_exists(self.SPLITTING_STAGE)
            reducing_done = self.done_file_exists(self.REDUCING_STAGE)

            if reducing_done and (not mapping_done or not splitting_done):
                raise ValueError("mapping and splitting must be complete before reducing")
            if splitting_done and not mapping_done:
                raise ValueError("mapping must be complete before splitting")
            step_progress.update(1)

            ## Figure out which stages we should run, based on requested `run_stages`
            self.should_run_mapping = not mapping_done
            self.should_run_splitting = not splitting_done
            self.should_run_reducing = not reducing_done
            self.should_run_finishing = True

            if run_stages:
                self.should_run_mapping &= self.MAPPING_STAGE in run_stages
                self.should_run_splitting &= self.SPLITTING_STAGE in run_stages
                self.should_run_reducing &= self.REDUCING_STAGE in run_stages
                self.should_run_finishing = "finishing" in run_stages

            ## Validate that we're operating on the same file set as the previous instance.
            self.input_paths = self.check_original_input_paths(self.input_paths)
            step_progress.update(1)

            ## Gather keys for execution.
            if self.should_run_mapping:
                self.map_files = self.get_remaining_map_keys()
                file_io.make_directory(
                    import_io.append_paths_to_pointer(self.tmp_path, self.ROW_COUNT_HISTOGRAMS_DIR),
                    exist_ok=True,
                )
                # If using mem_size thresholding, gather those keys too.
                if self.threshold_mode == "mem_size":
                    self.get_remaining_map_keys(which_histogram="mem_size")
                    file_io.make_directory(
                        import_io.append_paths_to_pointer(self.tmp_path, self.MEM_SIZE_HISTOGRAMS_DIR),
                        exist_ok=True,
                    )
            if self.should_run_splitting:
                if not (mapping_done or self.should_run_mapping):
                    raise ValueError("mapping must be complete before splitting")

                self.split_keys = self.get_remaining_split_keys()
                file_io.make_directory(
                    import_io.append_paths_to_pointer(self.tmp_path, self.SPLITTING_STAGE),
                    exist_ok=True,
                )
            if self.should_run_reducing:
                ## We don't pre-gather the plan for the reducing keys.
                ## It requires the full destination pixel map.
                if not (splitting_done or self.should_run_splitting):
                    raise ValueError("splitting must be complete before reducing")

                file_io.make_directory(
                    import_io.append_paths_to_pointer(self.tmp_path, self.REDUCING_STAGE),
                    exist_ok=True,
                )
            step_progress.update(1)

    def get_remaining_map_keys(self, which_histogram: str = "row_count"):
        """Gather remaining keys, dropping successful mapping tasks from histogram names.

        Args:
            which_histogram (str): Which histogram to check for completed tasks, either 'row_count'
                or 'mem_size'. Defaults to 'row_count'.

        Returns:
            list of tuple: The mapping keys *not* found in files like /resume/path/mapping_key.npz,
                along with their corresponding input file paths.

        Raises:
            ValueError: If `which_histogram` is not recognized, or if which_histogram is
                'mem_size' but the threshold_mode is not 'mem_size'.
        """
        if which_histogram == "row_count":
            prefix = file_io.get_upath(self.tmp_path) / self.ROW_COUNT_HISTOGRAMS_DIR
        elif which_histogram == "mem_size" and self.threshold_mode == "mem_size":
            prefix = file_io.get_upath(self.tmp_path) / self.MEM_SIZE_HISTOGRAMS_DIR
        elif which_histogram == "mem_size":
            raise ValueError("Cannot get remaining mem_size map keys when threshold_mode is not 'mem_size'.")
        else:
            raise ValueError(f"Unrecognized which_histogram value: {which_histogram}")

        map_file_pattern = re.compile(r"map_(\d+).npz")
        done_indexes = [
            int(match.group(1))
            for path in prefix.glob("*.npz")
            if (match := map_file_pattern.match(path.name))
        ]
        remaining_indexes = list(set(range(0, len(self.input_paths))) - (set(done_indexes)))
        return [(f"map_{key}", self.input_paths[key]) for key in remaining_indexes]

    def read_histogram(self, healpix_order, which_histogram: str = "row_count"):
        """Returns a histogram with the specified Healpix order's shape.

        This method attempts the following steps in order:
        1. Tries to locate and return a combined histogram.
        2. If a combined histogram is unavailable, combines partial histograms to create one.
        3. If no partial histograms are found, returns an empty histogram.

        Args:
            healpix_order (int): The desired Healpix order for the histogram.
            which_histogram (str): Which histogram to read, either "row_count" or "mem_size".
                Defaults to "row_count".


        Returns:
            numpy.ndarray: A one-dimensional array representing the histogram with the
                specified Healpix order.

        Raises:
            RuntimeError: If there are incomplete mapping stages.
            ValueError: If the histogram from the previous execution is incompatible with
                the highest Healpix order, or if `which_histogram` is invalid.
        """
        if which_histogram == "row_count":
            histogram_binary_file = self.ROW_COUNT_HISTOGRAM_BINARY_FILE
            histogram_directory = self.ROW_COUNT_HISTOGRAMS_DIR
        elif which_histogram == "mem_size" and self.threshold_mode == "mem_size":
            histogram_binary_file = self.MEM_SIZE_HISTOGRAM_BINARY_FILE
            histogram_directory = self.MEM_SIZE_HISTOGRAMS_DIR
        elif which_histogram == "mem_size":
            raise ValueError("Cannot read mem_size histogram when threshold_mode is not 'mem_size'.")
        else:
            raise ValueError(f"Unrecognized which_histogram value: {which_histogram}")

        file_name = import_io.append_paths_to_pointer(self.tmp_path, histogram_binary_file)

        # If no file, read the histogram from partial histograms and combine.
        if not file_name.exists():
            remaining_map_files = self.get_remaining_map_keys(which_histogram=which_histogram)
            if len(remaining_map_files) > 0:
                raise RuntimeError(f"{len(remaining_map_files)} map stages did not complete successfully.")
            histogram_files = import_io.find_files_matching_path(self.tmp_path, histogram_directory, "*.npz")
            aggregate_histogram = HistogramAggregator(healpix_order)
            for partial_file_name in histogram_files:
                partial = SparseHistogram.from_file(partial_file_name)
                aggregate_histogram.add(partial)

            file_name = import_io.append_paths_to_pointer(self.tmp_path, histogram_binary_file)
            with file_name.open("wb+") as file_handle:
                file_handle.write(aggregate_histogram.full_histogram)
            if self.delete_resume_log_files:
                file_io.remove_directory(
                    import_io.append_paths_to_pointer(self.tmp_path, histogram_directory),
                    ignore_errors=True,
                )

        with file_name.open("rb") as file_handle:
            full_histogram = frombuffer(file_handle.read(), dtype=np.int64)

        if len(full_histogram) != hp.order2npix(healpix_order):
            raise ValueError(
                "The histogram from the previous execution is incompatible with "
                + "the highest healpix order. To start the importing pipeline "
                + "from scratch with the current order set `resume` to False."
            )
        return full_histogram

    @classmethod
    def partial_histogram_file(cls, tmp_path, mapping_key: str, which_histogram: str = "row_count"):
        """File name for writing a histogram file to a special intermediate directory.

        As a side effect, this method may create the special intermediate directory.

        Args:
            tmp_path (str): where to write intermediate resume files.
            mapping_key (str): unique string for each mapping task (e.g. "map_57")
            which_histogram (str): which histogram to write, either "row_count" or "mem_size".
                Defaults to "row_count".

        Returns:
            str: Full path to the partial histogram file.
        """
        if which_histogram == "row_count":
            histograms_dir = cls.ROW_COUNT_HISTOGRAMS_DIR
        elif which_histogram == "mem_size":
            histograms_dir = cls.MEM_SIZE_HISTOGRAMS_DIR
        else:
            raise ValueError(f"Unrecognized which_histogram value: {which_histogram}")

        file_io.make_directory(
            import_io.append_paths_to_pointer(tmp_path, histograms_dir),
            exist_ok=True,
        )
        return import_io.append_paths_to_pointer(tmp_path, histograms_dir, f"{mapping_key}.npz")

    def get_remaining_split_keys(self):
        """Gather remaining keys, dropping successful split tasks from done file names.

        Returns:
            list of splitting keys *not* found in files like /resume/path/split_key.done
        """
        prefix = file_io.get_upath(self.tmp_path) / self.SPLITTING_STAGE
        split_file_pattern = re.compile(r"split_(\d+)_done")
        done_indexes = [int(split_file_pattern.match(path.name).group(1)) for path in prefix.glob("*_done")]
        remaining_indexes = list(set(range(0, len(self.input_paths))) - set(done_indexes))
        return [(f"split_{key}", self.input_paths[key]) for key in remaining_indexes]

    @classmethod
    def splitting_key_done(cls, tmp_path, splitting_key: str):
        """Mark a single splitting task as done

        Args:
            tmp_path (str): where to write intermediate resume files.
            splitting_key (str): unique string for each splitting task (e.g. "split_57")
        """
        cls.touch_key_done_file(tmp_path, cls.SPLITTING_STAGE, splitting_key)

    @classmethod
    def reducing_key_done(cls, tmp_path, reducing_key: str):
        """Mark a single reducing task as done

        Args:
            tmp_path (str): where to write intermediate resume files.
            reducing_key (str): unique string for each reducing task (e.g. "3_57")
        """
        cls.touch_key_done_file(tmp_path, cls.REDUCING_STAGE, reducing_key)

    def wait_for_mapping(self, futures):
        """Wait for mapping futures to complete."""
        self.wait_for_futures(futures, self.MAPPING_STAGE)
        remaining_map_items = self.get_remaining_map_keys()
        if len(remaining_map_items) > 0:
            raise RuntimeError("some map stages did not complete successfully.")
        self.touch_stage_done_file(self.MAPPING_STAGE)

    def get_alignment_file(
        self,
        raw_histogram,
        constant_healpix_order,
        highest_healpix_order,
        lowest_healpix_order,
        pixel_threshold,
        drop_empty_siblings,
        expected_total_rows,
        existing_pixels=None,
        raw_histogram_mem_size=None,
    ) -> UPath:
        """Get a pointer to the existing alignment file for the pipeline, or
        generate a new alignment using provided arguments.

        Args:
            raw_histogram (:obj:`np.array`): one-dimensional numpy array of long integers where the
                value at each index corresponds to the number of objects found at the healpix pixel.
            constant_healpix_order (int): if positive, use this as the order for
                all non-empty partitions. else, use remaining arguments.
            highest_healpix_order (int):  the highest healpix order (e.g. 5-10)
            lowest_healpix_order (int): the lowest healpix order (e.g. 1-5). specifying a lowest order
                constrains the partitioning to prevent spatially large pixels.
            pixel_threshold (int): the maximum number of objects allowed in a single pixel
            drop_empty_siblings (bool):  if 3 of 4 pixels are empty, keep only the non-empty pixel
            expected_total_rows (int): number of expected rows found in the dataset.
            existing_pixels (Sequence[tuple[int,int]]): the HEALPix pixels to include in the alignment
            raw_histogram_mem_size (:obj:`np.array`): one-dimensional numpy array of long integers
                where the value at each index corresponds to the memory size in bytes of objects
                found at the healpix pixel. Only required if threshold_mode is 'mem_size'.

        Returns:
            path to cached alignment file.
        """
        file_name = import_io.append_paths_to_pointer(self.tmp_path, self.ALIGNMENT_FILE)
        if not file_name.exists():
            # If existing_pixels, create an incremental alignment.
            if existing_pixels:
                alignment = pixel_math.generate_incremental_alignment(
                    raw_histogram,
                    existing_pixels=existing_pixels,
                    highest_order=highest_healpix_order,
                    lowest_order=lowest_healpix_order,
                    threshold=pixel_threshold,
                )
            # If constant_healpix_order is set, create a simple alignment.
            elif constant_healpix_order >= 0:
                alignment = self._generate_constant_healpix_order_alignment(
                    raw_histogram, constant_healpix_order
                )
            # Else, generate standard alignment based on thresholds.
            else:
                alignment = pixel_math.generate_alignment(
                    raw_histogram,
                    highest_order=highest_healpix_order,
                    lowest_order=lowest_healpix_order,
                    threshold=pixel_threshold,
                    drop_empty_siblings=drop_empty_siblings,
                    mem_size_histogram=raw_histogram_mem_size,
                )

            # Write alignment to file.
            with file_name.open("wb") as pickle_file:
                alignment = np.array([x if x is not None else [-1, -1, 0] for x in alignment], dtype=np.int64)
                pickle.dump(alignment, pickle_file)

        # Check that the destination pixel map (alignment file) matches expected total rows.
        if self.destination_pixel_map is None:
            with file_name.open("rb") as pickle_file:
                alignment = pickle.load(pickle_file)
            pixel_list = np.unique(alignment, axis=0)
            self.destination_pixel_map = {
                HealpixPixel(order, pix): row_count
                for (order, pix, row_count) in pixel_list
                if int(row_count) > 0
            }

        total_rows = sum(self.destination_pixel_map.values())
        if total_rows != expected_total_rows:
            raise ValueError(
                f"Number of rows ({total_rows}) does not match expectation ({expected_total_rows})"
            )

        return file_name

    def _generate_constant_healpix_order_alignment(self, raw_histogram_row_count, constant_healpix_order):
        """Generate alignment where all non-empty pixels are at the same healpix order.

        Args:
            raw_histogram_row_count (:obj:`np.array`): one-dimensional numpy array of long integers
                where the value at each index corresponds to the number of objects found at the
                healpix pixel.
            constant_healpix_order (int): the healpix order to assign to all non-empty pixels.

        Returns:
            np.array: alignment array where each entry is [order, pixel, row_count] or
                [order, pixel, row_count, mem_size] depending on threshold_mode.
        """
        alignment = np.full((len(raw_histogram_row_count), 3), [-1, -1, 0])
        for pixel_num, pixel_row_count_sum in enumerate(raw_histogram_row_count):
            alignment[pixel_num] = [
                constant_healpix_order,
                pixel_num,
                pixel_row_count_sum,
            ]
        return alignment

    def wait_for_splitting(self, futures):
        """Wait for splitting futures to complete."""
        self.wait_for_futures(futures, self.SPLITTING_STAGE)
        remaining_split_items = self.get_remaining_split_keys()
        if len(remaining_split_items) > 0:
            raise RuntimeError(f"{len(remaining_split_items)} split stages did not complete successfully.")
        self.touch_stage_done_file(self.SPLITTING_STAGE)

    def get_reduce_items(self):
        """Fetch a triple for each partition to reduce.

        Triple contains:

        - destination pixel (healpix pixel with both order and pixel)
        - number of rows expected for this pixel
        - reduce key (string of destination order+pixel)
        """
        if self.destination_pixel_map is None:
            raise RuntimeError("destination pixel map not provided for progress tracking.")

        reduced_pixels = self.read_done_pixels(self.REDUCING_STAGE)

        remaining_pixels = list(set(self.destination_pixel_map.keys()) - set(reduced_pixels))
        return [
            (hp_pixel, self.destination_pixel_map[hp_pixel], f"{hp_pixel.order}_{hp_pixel.pixel}")
            for hp_pixel in remaining_pixels
        ]

    def get_destination_pixels(self):
        """Create HealpixPixel list of all destination pixels."""
        if self.destination_pixel_map is None:
            raise RuntimeError("destination pixel map not known.")
        return list(self.destination_pixel_map.keys())

    def wait_for_reducing(self, futures):
        """Wait for reducing futures to complete."""
        self.wait_for_futures(futures, self.REDUCING_STAGE, fail_fast=True)
        remaining_reduce_items = self.get_reduce_items()
        if len(remaining_reduce_items) > 0:
            raise RuntimeError(f"{len(remaining_reduce_items)} reduce stages did not complete successfully.")
        self.touch_stage_done_file(self.REDUCING_STAGE)



from dataclasses import dataclass
from pathlib import Path

import hats
from hats.catalog import AssociationCatalog
from hats.io.file_io import get_upath
from typing_extensions import Self
from upath import UPath

from hats_import import ImportArguments


@dataclass
class AssociationArguments(ImportArguments):
    """Container class for holding arguments for partitioning association HATS catalogs."""

    catalog_type: str = "association"
    """level of catalog data, a mapping between catalogs"""
    allowed_catalog_types: tuple[str] = ("association",)
    """possible types of catalog to import with `AssociationArguments`"""
    should_write_skymap: bool = False
    """association catalogs do not contain skymap fits files"""

    @classmethod
    def reimport_from_hats(cls, path: str | Path | UPath, output_dir: str | Path | UPath, **kwargs) -> Self:
        """Generate the import arguments to reimport a HATS association with different parameters

        Args:
            path (str | Path | UPath): the path to the existing association catalog to reimport
            output_dir (str | Path | UPath): the path to output the reimported catalog to
            kwargs: any import arguments to update from the existing catalog. Can be any argument
                usually passed to :func:`AssociationArguments`

        Returns:
            A AssociationArguments object with the arguments from the existing association,
            and any updates from kwargs
        """
        catalog = hats.read_hats(get_upath(path))

        if not isinstance(catalog, AssociationCatalog):
            raise ValueError("The catalog must be of type `association`")

        addl_hats_properties = {
            "primary_catalog": catalog.catalog_info.primary_catalog,
            "primary_column": catalog.catalog_info.primary_column,
            "primary_column_association": catalog.catalog_info.primary_column_association,
            "join_catalog": catalog.catalog_info.join_catalog,
            "join_column": catalog.catalog_info.join_column,
            "join_column_association": catalog.catalog_info.join_column_association,
            "assn_max_separation": catalog.catalog_info.assn_max_separation,
            "contains_leaf_files": catalog.catalog_info.contains_leaf_files,
        }
        return super().reimport_from_hats(
            path, output_dir, addl_hats_properties=addl_hats_properties, **kwargs
        )



"""Tests of map reduce operations"""

import os

import numpy.testing as npt
import pandas as pd
import pyarrow.parquet as pq
import pytest
from hats import read_hats
from hats.io import get_parquet_metadata_pointer, paths
from hats.io.file_io import read_parquet_metadata
from hats.pixel_math.healpix_pixel import HealpixPixel

import hats_import.margin_cache.margin_cache as mc
from hats_import.margin_cache.margin_cache_arguments import MarginCacheArguments


@pytest.mark.dask(timeout=150)
@pytest.mark.parametrize(
    "small_sky_source_catalog",
    ["small_sky_source_catalog", "small_sky_source_npix_dir_catalog"],
    indirect=True,
)
def test_margin_cache_gen(small_sky_source_catalog, tmp_path, dask_client):
    """Test that margin cache generation works end to end."""
    args = MarginCacheArguments(
        margin_threshold=180.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=8,
        progress_bar=False,
        npix_suffix=".pq",
    )

    assert args.catalog.catalog_info.ra_column == "source_ra"

    mc.generate_margin_cache(args, dask_client)

    norder = 1
    npix = 47

    test_file = paths.pixel_catalog_file(args.catalog_path, HealpixPixel(norder, npix), npix_suffix=".pq")

    data = pd.read_parquet(test_file)

    assert len(data) == 88

    npt.assert_array_equal(
        data.columns,
        [
            "_healpix_29",
            "source_id",
            "source_ra",
            "source_dec",
            "mjd",
            "mag",
            "band",
            "object_id",
            "object_ra",
            "object_dec",
        ],
    )

    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.npix_suffix == ".pq"

    # Check that the data thumbnail does not exist. It should only exist for
    # main object/source catalogs.
    assert not (args.catalog_path / "dataset" / "data_thumbnail.parquet").exists()

    metadata = pq.read_metadata(test_file)
    assert metadata.num_row_groups == 1
    assert metadata.row_group(0).column(0).compression == "ZSTD"


@pytest.mark.dask(timeout=150)
def test_margin_cache_gen_negative_pixels(small_sky_source_catalog, tmp_path, dask_client):
    """Test that margin cache generation can generate a file for a negative pixel."""
    args = MarginCacheArguments(
        margin_threshold=3600.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=3,
        progress_bar=False,
        fine_filtering=False,
    )

    assert args.catalog.catalog_info.ra_column == "source_ra"

    mc.generate_margin_cache(args, dask_client)

    norder = 0
    npix = 7

    negative_test_file = paths.pixel_catalog_file(args.catalog_path, HealpixPixel(norder, npix))

    negative_data = pd.read_parquet(negative_test_file)

    assert len(negative_data) > 0


@pytest.mark.dask(timeout=150)
def test_generate_empty_margin_catalog(small_sky_object_catalog, tmp_path, dask_client):
    """Test that margin cache generation works with empty catalogs."""
    args = MarginCacheArguments(
        margin_threshold=10.0,
        input_catalog_path=small_sky_object_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        progress_bar=False,
    )

    mc.generate_margin_cache(args, dask_client)

    # Verify that an empty catalog was created with the correct metadata
    catalog = read_hats(args.catalog_path)
    object_cat = read_hats(small_sky_object_catalog)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert catalog.catalog_info.total_rows == 0
    assert catalog.catalog_info.hats_order == args.catalog.catalog_info.hats_order
    assert len(catalog.get_healpix_pixels()) == 0
    assert len(catalog.pixel_tree) == 0
    assert catalog.catalog_info.ra_column == "ra"
    assert catalog.catalog_info.dec_column == "dec"
    assert catalog.catalog_info.margin_threshold == 10.0
    assert catalog.schema == object_cat.schema

    # Check that the metadata files exist
    assert (args.catalog_path / "partition_info.csv").exists()
    assert (args.catalog_path / "hats.properties").exists()

    metadata_path = args.catalog_path / "dataset" / "_metadata"
    common_metadata_path = args.catalog_path / "dataset" / "_common_metadata"
    assert metadata_path.exists()
    assert common_metadata_path.exists()

    # Verify that the catalog contains no data files
    data_files = [
        f
        for f in (args.catalog_path / "dataset").rglob("*")
        if f.is_file() and f not in (metadata_path, common_metadata_path)
    ]
    assert len(list(data_files)) == 0

    # Check both pyarrow metadata files are correct
    metadata = read_parquet_metadata(metadata_path)
    assert metadata.num_rows == 0
    assert metadata.num_row_groups == 0
    assert metadata.schema.to_arrow_schema() == object_cat.schema

    common_metadata = read_parquet_metadata(common_metadata_path)
    assert common_metadata.num_rows == 0
    assert common_metadata.num_row_groups == 0
    assert common_metadata.schema.to_arrow_schema() == object_cat.schema


@pytest.mark.dask(timeout=150)
def test_margin_gen_nested_catalog(small_sky_nested_catalog, tmp_path, dask_client):
    args = MarginCacheArguments(
        margin_order=8,
        input_catalog_path=small_sky_nested_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        progress_bar=False,
    )

    mc.generate_margin_cache(args, dask_client)

    metadata_path = get_parquet_metadata_pointer(args.catalog_path)
    metadata = read_parquet_metadata(metadata_path)
    assert metadata.num_rows == 13

    norder = 2
    npix = 178

    test_file = paths.pixel_catalog_file(args.catalog_path, HealpixPixel(norder, npix))

    data = pd.read_parquet(test_file)

    assert len(data) == 2

    npt.assert_array_equal(
        data.columns,
        [
            "id",
            "ra",
            "dec",
            "ra_error",
            "dec_error",
            "lc",
            "_healpix_29",
        ],
    )

    catalog = read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path


@pytest.mark.dask(timeout=150)
def test_margin_cache_omit_metadata(small_sky_source_catalog, tmp_path, dask_client):
    args = MarginCacheArguments(
        margin_threshold=180.0,
        input_catalog_path=small_sky_source_catalog,
        output_path=tmp_path,
        output_artifact_name="catalog_cache",
        margin_order=8,
        progress_bar=False,
        create_metadata=False,
    )
    mc.generate_margin_cache(args, dask_client)

    # Check that the catalog was made correctly, but the metadata file does NOT exist
    catalog = read_hats(args.catalog_path)
    assert catalog.catalog_info.total_rows == 1717
    assert catalog.on_disk
    metadata_filename = os.path.join(args.catalog_path, "dataset", "_metadata")
    assert not os.path.exists(metadata_filename)
    assert catalog.catalog_path == args.catalog_path



"""Test resume file operations"""

from pathlib import Path

import numpy.testing as npt
import pytest

from hats_import.pipeline_resume_plan import PipelineResumePlan, get_formatted_stage_name


def test_done_file(tmp_path):
    """Verify expected behavior of done file"""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False)
    done_file = "action_done"

    ## done file should not exist
    assert not plan.done_file_exists(done_file)

    ## mark done and check that we can find it
    plan.touch_stage_done_file(done_file)
    assert plan.done_file_exists(done_file)

    ## mark done is idempotent
    plan.touch_stage_done_file(done_file)
    assert plan.done_file_exists(done_file)

    ## clear all the resume files and verify done file is gone.
    plan.clean_resume_files()

    assert not plan.done_file_exists(done_file)


def test_safe_to_resume(tmp_path):
    """Check that we throw errors when it's not safe to resume."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)
    plan.safe_to_resume()

    ## check is idempotent - intermediate directory exists but does not
    ## contain files.
    plan.safe_to_resume()

    ## take some action and verify that we can no longer resume without
    ## explicitly setting resume=True
    done_file = "action_done"
    plan.touch_stage_done_file(done_file)
    plan.safe_to_resume()

    ## If we mark as a resuming pipeline, we're safe to resume.
    resuming_plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=True)
    resuming_plan.safe_to_resume()

    ## If there are no more intermediate files, we don't need to set resume.
    plan.clean_resume_files()
    plan.safe_to_resume()


@pytest.mark.dask
def test_wait_for_futures(tmp_path, dask_client):
    """Test that we can wait around for futures to complete.

    Additionally test that relevant parts of the traceback are printed to stdout."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)

    def error_on_even(argument):
        """Silly little method used to test futures that fail under predictable conditions"""
        if argument % 2 == 0:
            raise RuntimeError("we are at odds with evens")

    ## Everything is fine if we're all odd
    futures = [dask_client.submit(error_on_even, 1)]
    plan.wait_for_futures(futures, "test")

    ## Throw an even in the mix, and we'll see some stages fail. Should cause whole stage to fail.
    futures = [dask_client.submit(error_on_even, 1), dask_client.submit(error_on_even, 2)]
    with pytest.raises(RuntimeError, match="Some test stages failed"):
        plan.wait_for_futures(futures, "test")


@pytest.mark.dask
def test_wait_for_futures_progress(tmp_path, dask_client, capsys):
    """Test that we can wait around for futures to complete.

    Additionally test that relevant parts of the traceback are printed to stdout."""
    plan = PipelineResumePlan(
        tmp_path=tmp_path,
        progress_bar=True,
        simple_progress_bar=True,
        resume=False,
        tqdm_kwargs={"ncols": 80},
    )

    def error_on_even(argument):
        """Silly little method used to test futures that fail under predictable conditions"""
        if argument % 2 == 0:
            raise RuntimeError("we are at odds with evens")

    ## Everything is fine if we're all odd, but use a silly name so it's
    ## clear that the stage name is present, and well-formatted.
    futures = [dask_client.submit(error_on_even, 1)]
    plan.wait_for_futures(futures, "teeeest")

    captured = capsys.readouterr()
    assert "Teeeest" in captured.err
    assert "100%" in captured.err


@pytest.mark.dask
def test_wait_for_futures_fail_fast(tmp_path, dask_client):
    """Test that we can wait around for futures to complete.

    Additionally test that relevant parts of the traceback are printed to stdout."""
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)

    def error_on_even(argument):
        """Silly little method used to test futures that fail under predictable conditions"""
        if argument % 2 == 0:
            raise RuntimeError("we are at odds with evens")

    futures = [dask_client.submit(error_on_even, 3), dask_client.submit(error_on_even, 4)]
    with pytest.raises(RuntimeError, match="we are at odds with evens"):
        plan.wait_for_futures(futures, "test", fail_fast=True)


def test_formatted_stage_name():
    """Test that we make pretty stage names for presenting in progress bars"""
    formatted = get_formatted_stage_name(None)
    assert formatted == "Progress  "

    formatted = get_formatted_stage_name("")
    assert formatted == "Progress  "

    formatted = get_formatted_stage_name("stage")
    assert formatted == "Stage     "

    formatted = get_formatted_stage_name("very long stage name")
    assert formatted == "Very long stage name"


def test_formatted_stage_name_with_pipeline():
    """Test that we make pretty stage names for presenting in progress bars"""
    formatted = get_formatted_stage_name(None, None)
    assert formatted == "Progress  "

    formatted = get_formatted_stage_name("", None)
    assert formatted == "Progress  "

    formatted = get_formatted_stage_name("stage", "pipeline")
    assert formatted == "Pipeline: Stage     "

    formatted = get_formatted_stage_name("very long stage name", "shorter pipeline")
    assert formatted == "Shorter pipeline: Very long stage name"


def test_check_original_input_paths(tmp_path, mixed_schema_csv_dir):
    plan = PipelineResumePlan(tmp_path=tmp_path, progress_bar=False, resume=False)

    input_file_list = [
        Path(mixed_schema_csv_dir) / "input_01.csv",
        Path(mixed_schema_csv_dir) / "input_02.csv",
    ]

    checked_files = plan.check_original_input_paths(input_file_list)

    round_trip_files = plan.check_original_input_paths(checked_files)

    npt.assert_array_equal(checked_files, round_trip_files)


def test_read_markers(test_data_dir):
    plan = PipelineResumePlan(tmp_path=test_data_dir / "markers", progress_bar=False)
    markers = plan.read_markers("mapping")
    assert markers == {"map_001": ["45"], "map_002": ["zippy"]}



"""test stuff."""

import hats
import numpy.testing as npt
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from hats.io.file_io import file_io
from pyarrow.parquet import ParquetFile

import hats_import.hipscat_conversion.run_conversion as runner
from hats_import.hipscat_conversion.arguments import ConversionArguments


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(TypeError, match="ConversionArguments"):
        runner.run(None, None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_artifact_name": "bad_arg_type"}
    with pytest.raises(TypeError, match="ConversionArguments"):
        runner.run(args, None)


# pylint: disable=unused-import
try:
    import healpy as hp

    HAVE_HEALPY = True
except ImportError:
    HAVE_HEALPY = False


@pytest.mark.skipif(not HAVE_HEALPY, reason="healpy is not installed")
@pytest.mark.dask
def test_run_conversion_object(
    test_data_dir,
    tmp_path,
    assert_parquet_file_ids,
    dask_client,
):
    """Test appropriate metadata is written"""

    input_catalog_dir = test_data_dir / "hipscat" / "small_sky_object_catalog"

    args = ConversionArguments(
        input_catalog_path=input_catalog_dir,
        output_path=tmp_path,
        output_artifact_name="small_sky_object_hats",
        progress_bar=False,
    )
    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = hats.read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_healpix_pixels()) == 1
    assert int(catalog.catalog_info.__pydantic_extra__["hats_estsize"]) > 0

    # Check that the catalog parquet file exists and contains correct object IDs
    output_file = args.catalog_path / "dataset" / "Norder=0" / "Dir=0" / "Npix=11.parquet"

    expected_ids = [*range(700, 831)]
    assert_parquet_file_ids(output_file, "id", expected_ids)

    # Check that the schema is correct for leaf parquet and _metadata files
    expected_parquet_schema = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
        ]
    )
    schema = pq.read_metadata(output_file).schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    assert schema.metadata is None
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    assert schema.metadata is None
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_common_metadata").schema.to_arrow_schema()
    assert schema.equals(expected_parquet_schema)
    assert schema.metadata is None

    data = file_io.read_parquet_file_to_pandas(
        output_file,
        columns=["id", "ra", "dec", "_healpix_29"],
    )
    assert "_healpix_29" in data.columns
    assert data.index.name is None

    # Check that the data thumbnail exists
    data_thumbnail_pointer = args.catalog_path / "dataset" / "data_thumbnail.parquet"
    assert data_thumbnail_pointer.exists()
    thumbnail = ParquetFile(data_thumbnail_pointer)
    thumbnail_schema = thumbnail.metadata.schema.to_arrow_schema()
    assert thumbnail_schema.equals(expected_parquet_schema)
    # The thumbnail has 1 row because the catalog has only 1 pixel
    assert len(thumbnail.read()) == 1


@pytest.mark.skipif(not HAVE_HEALPY, reason="healpy is not installed")
@pytest.mark.dask
def test_run_conversion_source(
    test_data_dir,
    tmp_path,
    dask_client,
):
    """Test appropriate metadata is written"""

    input_catalog_dir = test_data_dir / "hipscat" / "small_sky_source_catalog"

    args = ConversionArguments(
        input_catalog_path=input_catalog_dir,
        output_path=tmp_path,
        output_artifact_name="small_sky_source_hats",
        progress_bar=False,
    )
    runner.run(args, dask_client)

    # Check that the catalog metadata file exists
    catalog = hats.read_hats(args.catalog_path)
    assert catalog.on_disk
    assert catalog.catalog_path == args.catalog_path
    assert len(catalog.get_healpix_pixels()) == 14

    output_file = args.catalog_path / "dataset" / "Norder=2" / "Dir=0" / "Npix=185.parquet"

    source_columns = [
        "_healpix_29",
        "source_id",
        "source_ra",
        "source_dec",
        "mjd",
        "mag",
        "band",
        "object_id",
        "object_ra",
        "object_dec",
    ]
    schema = pq.read_metadata(output_file).schema
    npt.assert_array_equal(schema.names, source_columns)
    assert schema.to_arrow_schema().metadata is None
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_metadata").schema
    npt.assert_array_equal(schema.names, source_columns)
    assert schema.to_arrow_schema().metadata is None
    schema = pq.read_metadata(args.catalog_path / "dataset" / "_common_metadata").schema
    npt.assert_array_equal(schema.names, source_columns)
    assert schema.to_arrow_schema().metadata is None

    # Check that the data thumbnail exists
    data_thumbnail_pointer = args.catalog_path / "dataset" / "data_thumbnail.parquet"
    assert data_thumbnail_pointer.exists()
    thumbnail = ParquetFile(data_thumbnail_pointer)
    thumbnail_schema = thumbnail.metadata.schema.to_arrow_schema()
    assert thumbnail_schema.equals(schema.to_arrow_schema())
    # The thumbnail has 14 rows because the catalog has 14 pixels
    assert len(thumbnail.read()) == 14



"""Test catalog resume logic"""

from unittest.mock import MagicMock

import numpy as np
import numpy.testing as npt
import pytest
from hats.pixel_math.healpix_pixel import HealpixPixel
from hats.pixel_math.sparse_histogram import SparseHistogram

from hats_import.catalog.resume_plan import ResumePlan


def test_done_checks(tmp_path):
    """Verify that done files imply correct pipeline execution order:
    mapping > splitting > reducing
    """
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, resume=True)
    plan.touch_stage_done_file(ResumePlan.REDUCING_STAGE)

    with pytest.raises(ValueError, match="before reducing"):
        plan.gather_plan()

    plan.touch_stage_done_file(ResumePlan.SPLITTING_STAGE)
    with pytest.raises(ValueError, match="before reducing"):
        plan.gather_plan()

    plan.clean_resume_files()

    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, resume=True)
    plan.touch_stage_done_file(ResumePlan.SPLITTING_STAGE)
    with pytest.raises(ValueError, match="before splitting"):
        plan.gather_plan()


def test_same_input_paths(tmp_path, small_sky_single_file, formats_headers_csv):
    """Test that we can only resume if the input_paths are the same."""
    plan = ResumePlan(
        tmp_path=tmp_path,
        progress_bar=False,
        resume=True,
        input_paths=[small_sky_single_file, formats_headers_csv],
    )
    map_files = plan.map_files
    assert len(map_files) == 2

    with pytest.raises(ValueError, match="Different file set"):
        ResumePlan(
            tmp_path=tmp_path,
            progress_bar=False,
            resume=True,
            input_paths=[small_sky_single_file],
        )

    ## List is the same length, but includes a duplicate
    with pytest.raises(ValueError, match="Different file set"):
        ResumePlan(
            tmp_path=tmp_path,
            progress_bar=False,
            resume=True,
            input_paths=[small_sky_single_file, small_sky_single_file],
        )

    ## Includes a duplicate file, and we don't like that.
    with pytest.raises(ValueError, match="Different file set"):
        ResumePlan(
            tmp_path=tmp_path,
            progress_bar=False,
            resume=True,
            input_paths=[small_sky_single_file, small_sky_single_file, formats_headers_csv],
        )


def test_remaining_map_keys(tmp_path):
    """Test that we can read what we write into a histogram file."""
    num_inputs = 1_000
    input_paths = [f"foo_{i}" for i in range(0, num_inputs)]
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=input_paths)

    remaining_keys = plan.get_remaining_map_keys()
    assert len(remaining_keys) == num_inputs

    histogram = SparseHistogram([11], [131], 0)
    for i in range(0, num_inputs):
        histogram.to_file(ResumePlan.partial_histogram_file(tmp_path=tmp_path, mapping_key=f"map_{i}"))

    remaining_keys = plan.get_remaining_map_keys()
    assert len(remaining_keys) == 0


def test_read_write_histogram(tmp_path):
    """Test that we can read what we write into a histogram file."""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["foo1"])

    ## We're not ready to read the final histogram - missing partial histograms.
    with pytest.raises(RuntimeError, match="map stages"):
        result = plan.read_histogram(0)

    remaining_keys = plan.get_remaining_map_keys()
    assert remaining_keys == [("map_0", "foo1")]

    histogram = SparseHistogram([11], [131], 0)
    histogram.to_file(ResumePlan.partial_histogram_file(tmp_path=tmp_path, mapping_key="map_0"))

    remaining_keys = plan.get_remaining_map_keys()
    assert len(remaining_keys) == 0
    result = plan.read_histogram(0)
    npt.assert_array_equal(result, histogram.to_array())


def test_get_alignment_file(tmp_path):
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["foo1"])
    raw_histogram = np.full(12, 0)
    raw_histogram[11] = 131
    alignment_file = plan.get_alignment_file(raw_histogram, -1, 0, 0, 1_000, True, 131)

    alignment_file2 = plan.get_alignment_file(raw_histogram, -1, 0, 0, 1_000, True, 131)

    assert alignment_file == alignment_file2

    with pytest.raises(ValueError, match="does not match expectation"):
        plan.get_alignment_file(raw_histogram, -1, 0, 0, 1_000, True, 130)


def never_fails():
    """Method never fails, but never marks intermediate success file."""
    return


@pytest.mark.dask
def test_some_map_task_failures(tmp_path, dask_client):
    """Test that we only consider map stage successful if all partial files are written"""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["foo1"])

    ## Method doesn't FAIL, but it doesn't write out the partial histogram either.
    ## Since the intermediate files aren't found, we throw an error.
    futures = [dask_client.submit(never_fails)]
    with pytest.raises(RuntimeError, match="map stages"):
        plan.wait_for_mapping(futures)

    histogram = SparseHistogram([11], [131], 0)
    histogram.to_file(ResumePlan.partial_histogram_file(tmp_path=tmp_path, mapping_key="map_0"))

    ## Method succeeds, *and* partial histogram is present.
    futures = [dask_client.submit(never_fails)]
    plan.wait_for_mapping(futures)


def test_read_write_splitting_keys(tmp_path, small_sky_single_file, formats_headers_csv):
    """Test that we can read what we write into a reducing log file."""
    input_paths = [small_sky_single_file, formats_headers_csv]
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, resume=True, input_paths=input_paths)
    split_keys = plan.split_keys
    assert len(split_keys) == 2

    ResumePlan.touch_key_done_file(tmp_path, ResumePlan.SPLITTING_STAGE, "split_0")

    plan.gather_plan()
    split_keys = plan.split_keys
    assert len(split_keys) == 1
    assert split_keys[0][0] == "split_1"

    ResumePlan.touch_key_done_file(tmp_path, ResumePlan.SPLITTING_STAGE, "split_1")
    plan.gather_plan()
    split_keys = plan.split_keys
    assert len(split_keys) == 0

    plan.clean_resume_files()
    plan.gather_plan()
    split_keys = plan.split_keys
    assert len(split_keys) == 2


@pytest.mark.dask
def test_some_split_task_failures(tmp_path, dask_client):
    """Test that we only consider split stage successful if all done files are written"""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["foo1"])

    ## Method doesn't FAIL, but it doesn't write out the done file either.
    ## Since the intermediate files aren't found, we throw an error.
    futures = [dask_client.submit(never_fails)]
    with pytest.raises(RuntimeError, match="split stages"):
        plan.wait_for_splitting(futures)

    ResumePlan.touch_key_done_file(tmp_path, ResumePlan.SPLITTING_STAGE, "split_0")

    ## Method succeeds, and done file is present.
    futures = [dask_client.submit(never_fails)]
    plan.wait_for_splitting(futures)


def test_get_reduce_items(tmp_path):
    """Test generation of remaining reduce items"""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False)

    with pytest.raises(RuntimeError, match="destination pixel map"):
        remaining_reduce_items = plan.get_reduce_items()

    with pytest.raises(RuntimeError, match="destination pixel map"):
        remaining_reduce_items = plan.get_destination_pixels()

    plan.destination_pixel_map = {HealpixPixel(0, 11): 131}
    remaining_reduce_items = plan.get_reduce_items()
    assert len(remaining_reduce_items) == 1

    all_pixels = plan.get_destination_pixels()
    assert len(all_pixels) == 1

    ResumePlan.reducing_key_done(tmp_path=tmp_path, reducing_key="0_11")
    remaining_reduce_items = plan.get_reduce_items()
    assert len(remaining_reduce_items) == 0


@pytest.mark.dask
def test_some_reduce_task_failures(tmp_path, dask_client):
    """Test that we only consider reduce stage successful if all done files are written"""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False)

    plan.destination_pixel_map = {HealpixPixel(0, 11): 131}
    remaining_reduce_items = plan.get_reduce_items()
    assert len(remaining_reduce_items) == 1

    ## Method doesn't FAIL, but it doesn't write out the done file either.
    ## Since the intermediate files aren't found, we throw an error.
    futures = [dask_client.submit(never_fails)]
    with pytest.raises(RuntimeError, match="reduce stages"):
        plan.wait_for_reducing(futures)

    ResumePlan.touch_key_done_file(tmp_path, ResumePlan.REDUCING_STAGE, "0_11")

    ## Method succeeds, and done file is present.
    futures = [dask_client.submit(never_fails)]
    plan.wait_for_reducing(futures)


def test_run_stages(tmp_path):
    """Verify interaction with user-supplied run stages and done files"""
    shared_args = {
        "tmp_path": tmp_path / "stages",
        "progress_bar": False,
        "resume": True,
    }
    plan = ResumePlan(**shared_args)
    assert plan.should_run_mapping
    assert plan.should_run_splitting
    assert plan.should_run_reducing
    assert plan.should_run_finishing

    with pytest.raises(ValueError, match="before splitting"):
        plan.gather_plan(run_stages=["splitting"])

    plan.touch_stage_done_file(ResumePlan.MAPPING_STAGE)
    with pytest.raises(ValueError, match="before reducing"):
        plan.gather_plan(run_stages=["reducing"])

    plan.touch_stage_done_file(ResumePlan.SPLITTING_STAGE)
    plan.gather_plan()

    assert not plan.should_run_mapping
    assert not plan.should_run_splitting
    assert plan.should_run_reducing
    assert plan.should_run_finishing


def test_resume_plan_with_byte_pixel_threshold(tmp_path):
    """Test ResumePlan initialization with byte_pixel_threshold specified."""
    # Mock import_args with necessary attributes
    import_args = MagicMock()
    import_args.resume_kwargs_dict.return_value = {"tmp_path": tmp_path}
    import_args.debug_stats_only = False
    import_args.byte_pixel_threshold = 100  # Simulate a specified byte_pixel_threshold
    import_args.input_paths = ["file1", "file2"]

    # Initialize ResumePlan with the mocked import_args
    resume_plan = ResumePlan(import_args=import_args)

    # Assert that threshold_mode is set to "mem_size"
    assert resume_plan.threshold_mode == "mem_size"


def test_gather_plan_mem_size_mode(tmp_path):
    """Test gather_plan in mem_size mode."""
    # Mock import_args with necessary attributes
    import_args = MagicMock()
    import_args.resume_kwargs_dict.return_value = {"tmp_path": tmp_path}
    import_args.debug_stats_only = False
    import_args.byte_pixel_threshold = 100  # Simulate a specified byte_pixel_threshold
    import_args.input_paths = ["file1", "file2"]

    # Initialize ResumePlan with the mocked import_args
    plan = ResumePlan(import_args=import_args, tmp_path=tmp_path)

    # Mock methods to avoid actual file operations
    plan.get_remaining_map_keys = MagicMock(return_value=[("map_1", "file1"), ("map_2", "file2")])
    plan.done_file_exists = MagicMock(return_value=False)
    plan.check_original_input_paths = MagicMock(return_value=import_args.input_paths)
    plan.print_progress = MagicMock()

    # Call gather_plan and verify mem_size directory creation
    plan.gather_plan()
    plan.get_remaining_map_keys.assert_any_call(which_histogram="mem_size")


def test_get_remaining_map_keys_mem_size_and_invalid(tmp_path):
    """Test get_remaining_map_keys with mem_size and invalid options."""
    # Create a ResumePlan instance with mem_size threshold_mode
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["file1", "file2"])
    plan.threshold_mode = "mem_size"

    # Mock the directory structure to simulate mem_size histogram files
    mem_size_dir = tmp_path / plan.MEM_SIZE_HISTOGRAMS_DIR
    mem_size_dir.mkdir()
    (mem_size_dir / "map_0.npz").touch()

    # Call get_remaining_map_keys with mem_size
    remaining_keys = plan.get_remaining_map_keys(which_histogram="mem_size")
    assert len(remaining_keys) == 1
    assert remaining_keys[0] == ("map_1", "file2")

    # Call get_remaining_map_keys with an invalid option
    with pytest.raises(ValueError, match="Unrecognized which_histogram value"):
        plan.get_remaining_map_keys(which_histogram="invalid_option")

    # Raise error if threshold_mode is not mem_size
    plan.threshold_mode = "row_count"
    with pytest.raises(ValueError, match="when threshold_mode is not 'mem_size'"):
        plan.get_remaining_map_keys(which_histogram="mem_size")


def test_read_histogram_wrong_modes(tmp_path):
    """Test read_histogram with mem_size and invalid options."""
    plan = ResumePlan(tmp_path=tmp_path, progress_bar=False, input_paths=["file1", "file2"])
    assert plan.threshold_mode == "row_count"

    # Raise error for invalid which_histogram option when reading histogram
    with pytest.raises(ValueError, match="Cannot read mem_size histogram"):
        plan.read_histogram(2, which_histogram="mem_size")

    # Raise error for unrecognized which_histogram option
    with pytest.raises(ValueError, match="Unrecognized which_histogram value"):
        plan.read_histogram(2, which_histogram="invalid_option")



"""File reading generators for common file types."""

from .csv import CsvPyarrowReader, CsvReader, IndexedCsvReader
from .ecsv import AstropyEcsvReader
from .fits import FitsReader
from .input_reader import InputReader
from .parquet import IndexedParquetReader, ParquetPandasReader, ParquetPyarrowReader


def get_file_reader(
    file_format,
    chunksize=500_000,
    schema_file=None,
    column_names=None,
    skip_column_names=None,
    type_map=None,
    **kwargs,
):
    """Get a generator file reader for common file types

    Currently supported formats include:

    - ``"csv"``, comma separated values. may also be tab- or pipe-delimited
      includes `.csv.gz` and other compressed csv files
    - ``"fits"``, flexible image transport system. often used for astropy tables.
    - ``"parquet"``, compressed columnar data format
    - ``"ecsv"``, astropy's enhanced CSV
    - ``"indexed_csv"``, "index" style reader, that accepts a file with a list
      of csv files that are appended in-memory
    - ``"indexed_parquet"``, "index" style reader, that accepts a file with a list
      of parquet files that are appended in-memory

    Args:
        file_format (str): specifier for the file type and extension.
            If using an ``input_path`` argument, we will look for files with this string
            as the extension.
        chunksize (int): number of rows to read in a single iteration.
            for single-file readers, large files are split into batches based on this value.
            for index-style readers, we read files until we reach this chunksize and
            create a single batch in-memory.
        schema_file (str): path to a parquet schema file. if provided, header names
            and column types will be pulled from the parquet schema metadata.
        column_names (list[str]): for CSV files, the names of columns if no header
            is available. for fits files, a list of columns to *keep*.
        skip_column_names (list[str]): for fits files, a list of columns to remove.
        type_map (dict): for CSV files, the data types to use for columns
        kwargs: additional keyword arguments to pass to the underlying file reader.
    """
    if file_format == "csv":
        return CsvReader(
            chunksize=chunksize,
            schema_file=schema_file,
            column_names=column_names,
            type_map=type_map,
            **kwargs,
        )
    if file_format == "ecsv":
        return AstropyEcsvReader(**kwargs)
    if file_format == "fits":
        return FitsReader(
            chunksize=chunksize,
            column_names=column_names,
            skip_column_names=skip_column_names,
            **kwargs,
        )
    if file_format == "parquet":
        return ParquetPyarrowReader(chunksize=chunksize, **kwargs)
    if file_format == "indexed_csv":
        return IndexedCsvReader(
            chunksize=chunksize,
            schema_file=schema_file,
            column_names=column_names,
            type_map=type_map,
            **kwargs,
        )
    if file_format == "indexed_parquet":
        return IndexedParquetReader(chunksize=chunksize, **kwargs)
    raise NotImplementedError(f"File Format: {file_format} not supported")



"""Utility to hold the pipeline execution plan."""

from __future__ import annotations

from dataclasses import dataclass, field

import pandas as pd
from hats import pixel_math
from hats.io import file_io
from hats.pixel_math.healpix_pixel import HealpixPixel

import hats_import.file_io as import_io
from hats_import.margin_cache.margin_cache_arguments import MarginCacheArguments
from hats_import.pipeline_resume_plan import PipelineResumePlan


@dataclass
class MarginCachePlan(PipelineResumePlan):
    """Container class for holding the state of each file in the pipeline plan."""

    margin_pair_file: str | None = None
    partition_pixels: list[HealpixPixel] = field(default_factory=list)
    combined_pixels: list[HealpixPixel] = field(default_factory=list)

    MAPPING_STAGE = "mapping"
    REDUCING_STAGE = "reducing"
    MARGIN_PAIR_FILE = "margin_pair.csv"
    MAPPING_TOTAL_FILE = "mapping_total"

    def __init__(self, args: MarginCacheArguments):
        if not args.tmp_path:  # pragma: no cover (not reachable, but required for mypy)
            raise ValueError("tmp_path is required")
        super().__init__(pipeline_name="margin", **args.resume_kwargs_dict())
        self._gather_plan(args)

    def _gather_plan(self, args):
        """Initialize the plan."""
        with self.print_progress(total=3, stage_name="Planning") as step_progress:
            ## Make sure it's safe to use existing resume state.
            super().safe_to_resume()
            mapping_done = self.is_mapping_done()
            reducing_done = self.is_reducing_done()
            if reducing_done and (not mapping_done):
                raise ValueError("mapping must be complete before reducing")
            step_progress.update(1)

            self.partition_pixels = args.catalog.partition_info.get_healpix_pixels()
            negative_pixels = args.catalog.generate_negative_tree_pixels()
            self.combined_pixels = self.partition_pixels + negative_pixels
            self.margin_pair_file = import_io.append_paths_to_pointer(self.tmp_path, self.MARGIN_PAIR_FILE)
            if not self.margin_pair_file.exists():
                margin_pairs = _find_partition_margin_pixel_pairs(self.combined_pixels, args.margin_order)
                margin_pairs.to_csv(self.margin_pair_file, index=False)
            step_progress.update(1)

            file_io.make_directory(
                import_io.append_paths_to_pointer(self.tmp_path, self.MAPPING_STAGE),
                exist_ok=True,
            )
            file_io.make_directory(
                import_io.append_paths_to_pointer(self.tmp_path, self.REDUCING_STAGE),
                exist_ok=True,
            )

            step_progress.update(1)

    @classmethod
    def mapping_key_done(cls, tmp_path, mapping_key: str, num_rows: int):
        """Mark a single mapping task as done

        Args:
            tmp_path (str): where to write intermediate resume files.
            mapping_key (str): unique string for each mapping task (e.g. "map_1_24")
        """
        cls.write_marker_file(tmp_path, cls.MAPPING_STAGE, mapping_key, str(num_rows))

    def wait_for_mapping(self, futures):
        """Wait for mapping stage futures to complete."""
        self.wait_for_futures(futures, self.MAPPING_STAGE)
        remaining_pixels_to_map = self.get_remaining_map_keys()
        if len(remaining_pixels_to_map) > 0:
            raise RuntimeError(
                f"{len(remaining_pixels_to_map)} mapping stages did not complete successfully."
            )
        self.touch_stage_done_file(self.MAPPING_STAGE)

    def is_mapping_done(self) -> bool:
        """Are there sources left to count?"""
        return self.done_file_exists(self.MAPPING_STAGE)

    def get_mapping_total(self) -> int:
        """Find the total number of rows, based on the intermediate markers."""
        if not self.is_mapping_done():
            raise ValueError("mapping stage is not done yet.")

        total_marker_file = import_io.append_paths_to_pointer(self.tmp_path, self.MAPPING_TOTAL_FILE)

        if total_marker_file.exists():
            marker_value = file_io.load_text_file(total_marker_file)
            return _marker_value_to_int(marker_value)

        markers = self.read_markers(self.MAPPING_STAGE)
        total_marker_value = sum(_marker_value_to_int(value) for value in markers.values())
        file_io.write_string_to_file(total_marker_file, str(total_marker_value))
        return total_marker_value

    def get_remaining_map_keys(self):
        """Fetch a tuple for each pixel/partition left to map."""
        mapped_pixels = set(self.read_done_pixels(self.MAPPING_STAGE))
        remaining_pixels = list(set(self.partition_pixels) - set(mapped_pixels))
        return [(f"{hp_pixel.order}_{hp_pixel.pixel}", hp_pixel) for hp_pixel in remaining_pixels]

    @classmethod
    def reducing_key_done(cls, tmp_path, reducing_key: str):
        """Mark a single reducing task as done

        Args:
            tmp_path (str): where to write intermediate resume files.
            reducing_key (str): unique string for each reducing task (e.g. "3_57")
        """
        cls.touch_key_done_file(tmp_path, cls.REDUCING_STAGE, reducing_key)

    def get_remaining_reduce_keys(self):
        """Fetch a tuple for each object catalog pixel to reduce."""
        reduced_pixels = self.read_done_pixels(self.REDUCING_STAGE)
        remaining_pixels = list(set(self.combined_pixels) - set(reduced_pixels))
        return [(f"{hp_pixel.order}_{hp_pixel.pixel}", hp_pixel) for hp_pixel in remaining_pixels]

    def is_reducing_done(self) -> bool:
        """Are there partitions left to reduce?"""
        return self.done_file_exists(self.REDUCING_STAGE)

    def wait_for_reducing(self, futures):
        """Wait for reducing stage futures to complete."""
        self.wait_for_futures(futures, self.REDUCING_STAGE)
        remaining_sources_to_reduce = self.get_remaining_reduce_keys()
        if len(remaining_sources_to_reduce) > 0:
            raise RuntimeError(
                f"{len(remaining_sources_to_reduce)} reducing stages did not complete successfully."
            )
        self.touch_stage_done_file(self.REDUCING_STAGE)


def _find_partition_margin_pixel_pairs(combined_pixels, margin_order):
    """Creates a DataFrame filled with many-to-many connections between
    the catalog partition pixels and the negative margin pixels at `margin_order`.

    Args:
        combined_pixels (List[HealpixPixel]): union of catalog partition pixels
            and the negative tree pixels for the catalog
        margin_order (int): the order of healpixels that will be used to constrain
            the margin data before doing more precise boundary checking.
    """
    norders = []
    part_pix = []
    margin_pix = []

    for healpixel in combined_pixels:
        order = healpixel.order
        pix = healpixel.pixel

        d_order = margin_order - order

        margins = pixel_math.get_margin(order, pix, d_order)

        for m_p in margins:
            norders.append(order)
            part_pix.append(pix)
            margin_pix.append(m_p)

    margin_pairs_df = pd.DataFrame(
        zip(norders, part_pix, margin_pix),
        columns=["partition_order", "partition_pixel", "margin_pixel"],
    ).sort_values("margin_pixel")
    return margin_pairs_df


def _marker_value_to_int(marker_value: list[str]) -> int:
    """Convenience method to parse the contents of a marker file."""
    if len(marker_value) != 1:
        raise ValueError("Marker file should contain only one integer value.")
    return int(marker_value[0])



"""Data class to hold common runtime arguments for dataset creation."""

from __future__ import annotations

import re
from dataclasses import dataclass
from importlib.metadata import version
from pathlib import Path

import pandas as pd
from hats.catalog import TableProperties
from hats.io import file_io
from hats.io.validation import is_valid_catalog
from hats.pixel_math import spatial_index
from upath import UPath

import hats_import.file_io as import_io

# pylint: disable=too-many-instance-attributes


@dataclass
class RuntimeArguments:
    """Data class for holding runtime arguments"""

    ## Output
    output_path: str | Path | UPath | None = None
    """base path where new catalog should be output"""
    output_artifact_name: str = ""
    """short, convenient name for the catalog"""
    addl_hats_properties: dict | None = None
    """Any additional keyword arguments you would like to provide when writing
    the `hats.properties` file for the final HATS table. e.g. 
    {"hats_cols_default":"id, mjd", "hats_cols_survey_id":"unique_id", 
    "creator_did": "ivo://CDS/P/2MASS/J"}"""
    npix_suffix: str = ".parquet"
    """Suffix for pixel data. When specified as "/" each pixel will have a directory in its name."""
    npix_parquet_name: str | None = None
    """Name of the pixel parquet file to be used when npix_suffix=/. By default, it will be named
    after the pixel with a .parquet extension (e.g. 'Npix=10.parquet')"""
    write_table_kwargs: dict | None = None
    """additional keyword arguments to use when writing files to parquet (e.g. compression schemes)."""
    row_group_kwargs: dict | None = None
    """additional keyword arguments to use in creation of rowgroups when writing files to parquet."""
    should_write_skymap: bool = True
    """main catalogs should contain skymap fits files"""
    skymap_alt_orders: list[int] | None = None
    """Additional alternative healpix orders to write a HEALPix skymap."""
    create_thumbnail: bool = False
    """Create /dataset/data_thumbnail.parquet from one row of each data partition."""
    create_metadata: bool = True
    """Create /dataset/_metadata parquet from all data partitions."""

    ## Execution
    tmp_dir: str | Path | UPath | None = None
    """path for storing intermediate files"""
    resume: bool = True
    """If True, we try to read any existing intermediate files and continue to run
    the pipeline where we left off. If False, we start the import from scratch,
    overwriting any content of the output directory."""
    progress_bar: bool = True
    """if true, a progress bar will be displayed for user
    feedback of map reduce progress"""
    simple_progress_bar: bool = False
    """if displaying a progress bar, use a text-only simple progress
    bar instead of widget. this can be useful in some environments when running
    in a notebook where ipywidgets cannot be used (see `progress_bar` argument)"""
    tqdm_kwargs: dict | None = None
    """Additional arguments to pass to the tqdm progress bar. 
    See details https://tqdm.github.io/docs/tqdm/"""
    dask_tmp: str | Path | UPath | None = None
    """directory for dask worker space. this should be local to
    the execution of the pipeline, for speed of reads and writes"""
    dask_n_workers: int = 1
    """number of workers for the dask client"""
    dask_threads_per_worker: int = 1
    """number of threads per dask worker"""
    resume_tmp: str | Path | UPath | None = None
    """directory for intermediate resume files, when needed. see RTD for more info."""
    delete_intermediate_parquet_files: bool = True
    """should we delete the smaller intermediate parquet files generated in the
    splitting stage, once the relevant reducing stage is complete?"""
    delete_resume_log_files: bool = True
    """should we delete task-level done files once each stage is complete?
    if False, we will keep all done marker files at the end of the pipeline."""

    completion_email_address: str = ""
    """if provided, send an email to the indicated email address once the 
    import pipeline has completed."""

    catalog_path: UPath | None = None
    """constructed output path for the catalog that will be something like
    <output_path>/<output_artifact_name>"""
    tmp_path: UPath | None = None
    """constructed temp path - defaults to tmp_dir, then dask_tmp, but will create
    a new temp directory under catalog_path if no other options are provided"""
    tmp_base_path: UPath | None = None
    """temporary base directory: either `tmp_dir` or `dask_dir`, if those were provided by the user"""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        if not self.output_path:
            raise ValueError("output_path is required")
        if not self.output_artifact_name:
            raise ValueError("output_artifact_name is required")
        if re.search(r"[^A-Za-z0-9\._\-\\]", self.output_artifact_name):
            raise ValueError("output_artifact_name contains invalid characters")

        if self.dask_n_workers <= 0:
            raise ValueError("dask_n_workers should be greather than 0")
        if self.dask_threads_per_worker <= 0:
            raise ValueError("dask_threads_per_worker should be greater than 0")

        self.catalog_path = file_io.get_upath(self.output_path) / self.output_artifact_name
        if is_valid_catalog(self.catalog_path):
            raise ValueError(f"Output path {self.catalog_path} already contains a valid catalog")
        if not self.resume:
            file_io.remove_directory(self.catalog_path, ignore_errors=True)
        file_io.make_directory(self.catalog_path, exist_ok=True)

        if self.tmp_dir and str(self.tmp_dir) != str(self.output_path):
            self.tmp_path = import_io.append_paths_to_pointer(
                self.tmp_dir, self.output_artifact_name, "intermediate"
            )
            self.tmp_base_path = import_io.append_paths_to_pointer(self.tmp_dir, self.output_artifact_name)
        elif self.dask_tmp and str(self.dask_tmp) != str(self.output_path):
            self.tmp_path = import_io.append_paths_to_pointer(
                self.dask_tmp, self.output_artifact_name, "intermediate"
            )
            self.tmp_base_path = import_io.append_paths_to_pointer(self.dask_tmp, self.output_artifact_name)
        else:
            self.tmp_path = import_io.append_paths_to_pointer(self.catalog_path, "intermediate")
        if not self.resume:
            file_io.remove_directory(self.tmp_path, ignore_errors=True)
        file_io.make_directory(self.tmp_path, exist_ok=True)
        if self.resume_tmp:
            self.resume_tmp = import_io.append_paths_to_pointer(self.resume_tmp, self.output_artifact_name)
        else:
            self.resume_tmp = self.tmp_path

        if self.write_table_kwargs is None:
            self.write_table_kwargs = {}
        if "compression" not in self.write_table_kwargs:
            self.write_table_kwargs = self.write_table_kwargs | {
                "compression": "ZSTD",
                "compression_level": 15,
            }

    def extra_property_dict(self):
        """Generate additional HATS properties for this import run as a dictionary."""
        properties = {}
        if self.addl_hats_properties:
            properties = self.addl_hats_properties
        return TableProperties.new_provenance_dict(
            self.catalog_path, builder=f"hats-import v{version('hats-import')}", **properties
        )

    def resume_kwargs_dict(self):
        """Convenience method to convert fields for resume functionality."""
        return {
            "resume": self.resume,
            "progress_bar": self.progress_bar,
            "simple_progress_bar": self.simple_progress_bar,
            "tmp_path": self.resume_tmp,
            "tmp_base_path": self.tmp_base_path,
            "delete_resume_log_files": self.delete_resume_log_files,
            "delete_intermediate_parquet_files": self.delete_intermediate_parquet_files,
            "tqdm_kwargs": self.tqdm_kwargs,
        }


def find_input_paths(input_path="", file_matcher="", input_file_list=None):
    """Helper method to find input paths, given either a prefix and format, or an
    explicit list of paths.

    Args:
        input_path (str): prefix to search for
        file_matcher (str): matcher to use when searching for files
        input_file_list (List[str]): list of input paths
    Returns:
        matching files, if input_path is provided, otherwise, input_file_list
    Raises:
        FileNotFoundError: if no files are found at the input_path and the provided list is empty.
    """
    input_paths = []
    if input_path:
        if input_file_list:
            raise ValueError("exactly one of input_path or input_file_list is required")
        input_paths = import_io.find_files_matching_path(input_path, file_matcher)
    elif input_file_list is not None:
        if pd.api.types.is_list_like(input_file_list):
            # It's common for users to accidentally pass in an empty list. Give them a friendly error.
            if len(input_file_list) == 0:
                raise ValueError("input_file_list is empty")
        else:
            input_file_list = [input_file_list]
        input_paths = input_file_list
    else:
        raise ValueError("exactly one of input_path or input_file_list is required")
    if len(input_paths) == 0:
        raise FileNotFoundError("No input files found")
    return input_paths


def check_healpix_order_range(
    order, field_name, lower_bound=0, upper_bound=spatial_index.SPATIAL_INDEX_ORDER
):
    """Helper method to check if the ``order`` is within the range determined by the
    ``lower_bound`` and ``upper_bound``, inclusive.

    Args:
        order (int): healpix order to check
        field_name (str): field name to use in the error message
        lower_bound (int): lower bound of range
        upper_bound (int): upper bound of range
    Raise:
        ValueError: if the order is outside the specified range, or bounds
            are unreasonable.
    """
    if lower_bound < 0:
        raise ValueError("healpix orders must be positive")
    if upper_bound > spatial_index.SPATIAL_INDEX_ORDER:
        raise ValueError(f"healpix order should be <= {spatial_index.SPATIAL_INDEX_ORDER}")
    if not lower_bound <= order <= upper_bound:
        raise ValueError(f"{field_name} should be between {lower_bound} and {upper_bound}")



"""Test dataframe-generating file readers"""

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from hats.catalog import TableProperties

from hats_import.catalog.file_readers import (
    CsvPyarrowReader,
    CsvReader,
    FitsReader,
    IndexedCsvReader,
    IndexedParquetReader,
    ParquetPandasReader,
    ParquetPyarrowReader,
    get_file_reader,
)


# pylint: disable=redefined-outer-name
@pytest.fixture
def basic_catalog_info():
    info = {
        "catalog_name": "test_catalog",
        "catalog_type": "object",
        "total_rows": 100,
        "ra_column": "ra",
        "dec_column": "dec",
    }
    return TableProperties(**info)


def test_unknown_file_type():
    """File reader factory method should fail for unknown file types"""
    with pytest.raises(NotImplementedError):
        get_file_reader("")
    with pytest.raises(NotImplementedError):
        get_file_reader("unknown")


def test_file_exists():
    """File reader factory method should fail for missing files or directories"""
    with pytest.raises(FileNotFoundError, match="File not found"):
        next(CsvReader().read("foo_not_really_a_path"))


def test_csv_reader(small_sky_single_file):
    """Verify we can read the csv file into a single data frame."""
    total_chunks = 0
    for frame in CsvReader().read(small_sky_single_file):
        total_chunks += 1
        assert len(frame) == 131

    assert total_chunks == 1


def test_csv_gzip_reader(formats_dir):
    """Verify we can read the csv file into a single data frame."""
    total_chunks = 0
    for frame in CsvReader(compression="zip").read(formats_dir / "catalog.zip"):
        total_chunks += 1
        assert len(frame) == 131

    assert total_chunks == 1

    total_chunks = 0
    for frame in CsvReader(compression="gzip").read(formats_dir / "catalog.csv.gz"):
        total_chunks += 1
        assert len(frame) == 131

    assert total_chunks == 1


def test_csv_reader_chunks(small_sky_single_file):
    """Verify we can read the csv file into multiple data frames."""
    total_chunks = 0
    for frame in CsvReader(chunksize=15).read(small_sky_single_file):
        total_chunks += 1
        assert len(frame) <= 15

    assert total_chunks == 9

    total_chunks = 0
    for frame in CsvReader(chunksize=1).read(small_sky_single_file):
        total_chunks += 1
        assert len(frame) == 1

    assert total_chunks == 131


def test_csv_reader_no_headers(small_sky_single_file):
    """Verify we can read the csv file without a header row."""
    total_chunks = 0
    for frame in CsvReader(header=None).read(small_sky_single_file):
        total_chunks += 1
        # The "header" row is interpreted as just another data row
        assert len(frame) == 132

    assert total_chunks == 1


def test_csv_reader_parquet_metadata(small_sky_single_file, tmp_path):
    """Verify we can read the csv file without a header row."""
    small_sky_schema = pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.float64()),
            pa.field("dec_error", pa.float64()),
        ]
    )
    schema_file = tmp_path / "metadata.parquet"
    pq.write_metadata(
        small_sky_schema,
        schema_file,
    )

    frame = next(CsvReader(schema_file=schema_file).read(small_sky_single_file))
    assert len(frame) == 131

    column_types = frame.dtypes.to_dict()
    expected_column_types = {
        "id": pd.ArrowDtype(pa.int64()),
        "ra": pd.ArrowDtype(pa.float64()),
        "dec": pd.ArrowDtype(pa.float64()),
        "ra_error": pd.ArrowDtype(pa.float64()),
        "dec_error": pd.ArrowDtype(pa.float64()),
    }
    assert np.all(column_types == expected_column_types)


def test_csv_reader_kwargs(small_sky_single_file):
    """Verify we can read the csv file using kwargs passed to read_csv call."""

    ## Input file has 5 columns: ["id", "ra", "dec", "ra_error", "dec_error"]
    frame = next(CsvReader(usecols=["id", "ra", "dec"]).read(small_sky_single_file))
    assert len(frame) == 131

    assert len(frame.columns) == 3
    assert np.all(frame.columns == ["id", "ra", "dec"])


def test_csv_reader_pipe_delimited(formats_pipe_csv, tmp_path):
    """Verify we can read a pipe-delimited csv file without a header row."""
    total_chunks = 0
    for frame in CsvReader(header=None, sep="|").read(formats_pipe_csv):
        total_chunks += 1
        assert len(frame) == 3
        assert np.all(frame[0] == ["AA", "BB", "CC"])

    assert total_chunks == 1

    ## Provide header names and types in a few formats
    frame = next(
        CsvReader(
            column_names=["letters", "ints", "empty", "numeric"],
            header=None,
            sep="|",
        ).read(formats_pipe_csv)
    )
    assert len(frame) == 3
    assert np.all(frame["letters"] == ["AA", "BB", "CC"])
    column_types = frame.dtypes.to_dict()
    expected_column_types = {
        "letters": object,
        "ints": int,
        "empty": float,
        "numeric": float,
    }
    assert np.all(column_types == expected_column_types)

    frame = next(
        CsvReader(
            column_names=["letters", "ints", "empty", "numeric"],
            type_map={
                "letters": object,
                "ints": int,
                "empty": "Int64",
                "numeric": int,
            },
            header=None,
            sep="|",
        ).read(formats_pipe_csv)
    )
    assert len(frame) == 3
    assert np.all(frame["letters"] == ["AA", "BB", "CC"])
    column_types = frame.dtypes.to_dict()
    expected_column_types = {
        "letters": object,
        "ints": int,
        "empty": pd.Int64Dtype(),
        "numeric": int,
    }
    assert np.all(column_types == expected_column_types)

    parquet_schema_types = pa.schema(
        [
            pa.field("letters", pa.string()),
            pa.field("ints", pa.int64()),
            pa.field("empty", pa.int64()),
            pa.field("numeric", pa.int64()),
        ]
    )
    schema_file = tmp_path / "metadata.parquet"
    pq.write_metadata(parquet_schema_types, schema_file)

    frame = next(CsvReader(schema_file=schema_file, header=None, sep="|").read(formats_pipe_csv))

    assert len(frame) == 3
    assert np.all(frame["letters"] == ["AA", "BB", "CC"])
    column_types = frame.dtypes.to_dict()
    expected_column_types = {
        "letters": pd.ArrowDtype(pa.string()),
        "ints": pd.ArrowDtype(pa.int64()),
        "empty": pd.ArrowDtype(pa.int64()),
        "numeric": pd.ArrowDtype(pa.int64()),
    }
    assert np.all(column_types == expected_column_types)


def test_csv_pyarrow_reader_parquet_metadata(small_sky_single_file, tmp_path):
    """Verify we can read the csv file without a header row."""
    small_sky_schema = pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.float64()),
            pa.field("dec_error", pa.float64()),
        ]
    )
    schema_file = tmp_path / "metadata.parquet"
    pq.write_metadata(small_sky_schema, schema_file)

    frame = next(CsvPyarrowReader(schema_file=schema_file).read(small_sky_single_file))
    assert len(frame) == 131

    assert frame.schema.equals(small_sky_schema)


def test_indexed_csv_reader(indexed_files_dir):
    # Chunksize covers all the inputs.
    total_chunks = 0
    for frame in IndexedCsvReader(chunksize=10_000).read(indexed_files_dir / "csv_list_single.txt"):
        total_chunks += 1
        assert len(frame) == 131

    assert total_chunks == 1

    # Chunksize requires splitting into just a few batches.
    total_chunks = 0
    for frame in IndexedCsvReader(chunksize=60).read(indexed_files_dir / "csv_list_single.txt"):
        total_chunks += 1
        assert len(frame) < 60

    assert total_chunks == 3

    # Requesting a very small chunksize. This will split up reads on the CSV.
    total_chunks = 0
    for frame in IndexedCsvReader(chunksize=5).read(indexed_files_dir / "csv_list_single.txt"):
        total_chunks += 1
        assert len(frame) <= 5

    assert total_chunks == 29


def test_parquet_reader(parquet_shards_shard_44_0):
    """Verify we can read the parquet file into a single data frame."""
    # Test ParquetPandasReader.
    total_chunks = 0
    for frame in ParquetPandasReader().read(parquet_shards_shard_44_0):
        total_chunks += 1
        assert len(frame) == 7

    assert total_chunks == 1

    # Test ParquetPyarrowReader.
    total_chunks = 0
    for frame in ParquetPyarrowReader().read(parquet_shards_shard_44_0):
        total_chunks += 1
        assert len(frame) == 7

    assert total_chunks == 1


def test_parquet_reader_chunked(parquet_shards_shard_44_0):
    """Verify we can read the parquet file into a single data frame."""
    total_chunks = 0
    for frame in ParquetPyarrowReader(chunksize=1).read(parquet_shards_shard_44_0):
        total_chunks += 1
        assert len(frame) == 1
    assert total_chunks == 7


def test_parquet_reader_by_row_group(multi_row_group_parquet):
    """Verify we can read the parquet file into a single data frame, iterating
    by row group."""
    # Check number of row groups in test file.
    assert pq.ParquetFile(multi_row_group_parquet).num_row_groups == 8

    # Check we can iterate by row group, with ParquetPandasReader.
    total_row_groups = 0
    for row_group in ParquetPandasReader(iterate_by_row_groups=True).read(multi_row_group_parquet):
        total_row_groups += 1
        assert len(row_group) > 0
    assert total_row_groups == 8

    # Check we can iterate by row group, with ParquetPyarrowReader.
    total_row_groups = 0
    for row_group in ParquetPyarrowReader(iterate_by_row_groups=True).read(multi_row_group_parquet):
        total_row_groups += 1
        assert len(row_group) > 0
    assert total_row_groups == 8


def test_indexed_parquet_reader(indexed_files_dir):
    # Chunksize covers all the inputs.
    total_chunks = 0
    for frame in get_file_reader("indexed_parquet", chunksize=10_000).read(
        indexed_files_dir / "parquet_list_single.txt"
    ):
        total_chunks += 1
        assert len(frame) == 131

    assert total_chunks == 1

    # Chunksize requires splitting into just a few batches.
    total_chunks = 0
    for frame in IndexedParquetReader(chunksize=60).read(indexed_files_dir / "parquet_list_single.txt"):
        total_chunks += 1
        assert len(frame) < 60

    assert total_chunks == 3

    # Requesting a very small chunksize. This will split up reads on the CSV.
    total_chunks = 0
    for frame in IndexedParquetReader(chunksize=5).read(indexed_files_dir / "parquet_list_single.txt"):
        total_chunks += 1
        assert len(frame) <= 5

    assert total_chunks == 29


def test_parquet_reader_columns(parquet_shards_shard_44_0):
    """Verify we can read a subset of columns."""
    column_subset = ["id", "dec"]

    # test column_names class property
    for frame in ParquetPyarrowReader(column_names=column_subset).read(parquet_shards_shard_44_0):
        assert set(frame.column_names) == set(column_subset)

    # test read_columns kwarg
    for frame in ParquetPyarrowReader().read(parquet_shards_shard_44_0, read_columns=column_subset):
        assert set(frame.column_names) == set(column_subset)


def test_read_fits(formats_fits):
    """Success case - fits file that exists being read as fits"""
    total_chunks = 0
    for table in FitsReader().read(formats_fits):
        total_chunks += 1
        assert len(table) == 131

    assert total_chunks == 1


def test_read_fits_chunked(formats_fits):
    """Success case - fits file that exists being read as fits in chunks"""
    total_chunks = 0
    for table in FitsReader(chunksize=1).read(formats_fits):
        total_chunks += 1
        assert len(table) == 1

    assert total_chunks == 131


def test_read_fits_columns(formats_fits):
    """Success case - column filtering on reading fits file"""
    table = next(FitsReader(column_names=["id", "ra", "dec"]).read(formats_fits))
    assert list(table.column_names) == ["id", "ra", "dec"]

    table = next(FitsReader(column_names=["id", "ra", "dec"]).read(formats_fits, read_columns=["ra", "dec"]))
    assert list(table.column_names) == ["ra", "dec"]

    table = next(FitsReader(skip_column_names=["ra_error", "dec_error"]).read(formats_fits))
    assert list(table.column_names) == ["id", "ra", "dec", "test_id"]

    table = next(
        FitsReader(skip_column_names=["ra_error", "dec_error"]).read(formats_fits, read_columns=["ra", "dec"])
    )
    assert list(table.column_names) == ["ra", "dec"]


def test_read_fits_with_nested_columns(formats_fits_nested):
    """Check that nested array data are processed."""
    table = next(FitsReader().read(formats_fits_nested))
    assert table["COEFF"].type == pa.list_(pa.float64(), 10)


def test_read_fits_with_nested_tensor_columns(formats_fits_tensor):
    """Check that nested tensor data are processed."""
    table = next(FitsReader(flatten_tensors=False).read(formats_fits_tensor))
    assert table["FLUX"].type == pa.fixed_shape_tensor(pa.float32(), (5, 5))

    table = next(FitsReader(flatten_tensors=True).read(formats_fits_tensor))
    assert table["FLUX"].type == pa.list_(pa.float32(), 25)



"""Utility to hold all arguments required throughout partitioning"""

from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Sequence

import hats
from hats.catalog import TableProperties
from hats.catalog.catalog_collection import CatalogCollection
from hats.io.file_io import get_upath
from hats.io.paths import DATASET_DIR, HIVE_COLUMNS, PARTITION_ORDER
from hats.io.validation import is_valid_catalog
from hats.pixel_math.spatial_index import SPATIAL_INDEX_COLUMN, SPATIAL_INDEX_ORDER
from typing_extensions import Self
from upath import UPath

from hats_import.catalog.file_readers import InputReader, get_file_reader
from hats_import.catalog.file_readers.parquet import ParquetPyarrowReader
from hats_import.runtime_arguments import RuntimeArguments, check_healpix_order_range, find_input_paths

# pylint: disable=too-many-locals,too-many-arguments,too-many-instance-attributes,too-many-branches,too-few-public-methods


@dataclass
class ImportArguments(RuntimeArguments):
    """Container class for holding arguments for partitioning input
    data into a HATS catalog."""

    catalog_type: str = "object"
    """level of catalog data, object (things in the sky) or source (detections)"""
    allowed_catalog_types: tuple[str] = ("source", "object", "map")
    """possible types of catalog to import with `ImportArguments`"""
    input_path: str | Path | UPath | None = None
    """path to search for the input data"""
    input_file_list: list[str | Path | UPath] = field(default_factory=list)
    """can be used instead of input_path to import only specified files"""
    input_paths: list[str | Path | UPath] = field(default_factory=list)
    """resolved list of all files that will be used in the importer"""

    ra_column: str = "ra"
    """column for right ascension"""
    dec_column: str = "dec"
    """column for declination"""
    use_healpix_29: bool = False
    """use an existing healpix-based hats spatial index as the position, instead of ra/dec"""
    sort_columns: str | None = None
    """column for survey identifier, or other sortable column. if sorting by multiple columns,
    they should be comma-separated. If `add_healpix_29=True`, `_healpix_29` will be the primary sort key, 
    but the provided sorting will be used for any rows within the same higher-order pixel space."""
    add_healpix_29: bool = True
    """add the healpix-based hats spatial index field alongside the data"""

    use_schema_file: str | Path | UPath | None = None
    """path to a parquet file with schema metadata. this will be used for column
    metadata when writing the files, if specified"""
    expected_total_rows: int = 0
    """number of expected rows found in the dataset. if non-zero, and we find we have 
    a different number of rows, the pipeline will exit."""
    constant_healpix_order: int = -1
    """healpix order to use when mapping. if this is
    a positive number, this will be the order of all final pixels and we
    will not combine pixels according to the threshold"""
    lowest_healpix_order: int = 0
    """when determining bins for the final partitioning, the lowest possible healpix order 
    for resulting pixels. setting this higher than 0 will prevent creating
    partitions with a large area on the sky."""
    highest_healpix_order: int = 10
    """healpix order to use when mapping. this will
    not necessarily be the order used in the final catalog, as we may combine
    pixels that don't meed the threshold"""
    pixel_threshold: int = 1_000_000
    """when determining bins for the final partitioning, the maximum number 
    of rows for a single resulting pixel. we may combine hierarchically until 
    we near the ``pixel_threshold``"""
    byte_pixel_threshold: int | None = None
    """when determining bins for the final partitioning, the maximum number
    of rows for a single resulting pixel, expressed in bytes. we may combine hierarchically until
    we near the ``byte_pixel_threshold``. if this is set, it will override
    ``pixel_threshold``."""
    drop_empty_siblings: bool = True
    """when determining bins for the final partitioning, should we keep result pixels
    at a higher order (smaller area) if the 3 sibling pixels are empty. setting this to 
    False will result in the same number of result pixels, but they may differ in Norder"""
    mapping_healpix_order: int = -1
    """healpix order to use when mapping. will be
    ``highest_healpix_order`` unless a positive value is provided for
    ``constant_healpix_order``"""
    run_stages: list[str] = field(default_factory=list)
    """list of parallel stages to run. options are ['mapping', 'splitting', 'reducing',
    'finishing']. ['planning', 'binning'] stages are not optional.
    this can be used to force the pipeline to stop after an early stage, to allow the
    user to reset the dask client with different resources for different stages of
    the workflow. if not specified, we will run all pipeline stages."""
    debug_stats_only: bool = False
    """do not perform a map reduce and don't create a new
    catalog. generate the partition info"""
    file_reader: InputReader | str | None = None
    """instance of input reader that specifies arguments necessary for reading
    from your input files"""
    existing_pixels: Sequence[tuple[int, int]] | None = None
    """the list of HEALPix pixels to include in the alignment"""

    def __post_init__(self):
        self._check_arguments()

    def _check_arguments(self):
        """Check existence and consistency of argument values"""
        super()._check_arguments()

        if self.lowest_healpix_order == self.highest_healpix_order:
            self.constant_healpix_order = self.lowest_healpix_order
        if self.constant_healpix_order >= 0:
            check_healpix_order_range(self.constant_healpix_order, "constant_healpix_order")
            self.mapping_healpix_order = self.constant_healpix_order
        else:
            check_healpix_order_range(self.highest_healpix_order, "highest_healpix_order")
            check_healpix_order_range(
                self.lowest_healpix_order, "lowest_healpix_order", upper_bound=self.highest_healpix_order
            )
            if not isinstance(self.pixel_threshold, int) or self.pixel_threshold < 1:
                raise ValueError("pixel_threshold must be a positive integer")
            self.mapping_healpix_order = self.highest_healpix_order

        if self.existing_pixels:
            max_existing_order = max(p[0] for p in self.existing_pixels)
            if self.mapping_healpix_order < max_existing_order:
                raise ValueError("`highest_order` needs to be >= current `existing_pixels` order")

        if self.catalog_type not in self.allowed_catalog_types:
            raise ValueError(f"catalog_type should be one of {self.allowed_catalog_types}")

        if self.file_reader is None:
            raise ValueError("file_reader is required")
        if isinstance(self.file_reader, str):
            self.file_reader = get_file_reader(self.file_reader)

        if self.use_healpix_29:
            self.add_healpix_29 = False
            if self.sort_columns:
                raise ValueError("When using _healpix_29 for position, no sort columns should be added")

        # Validate byte_pixel_threshold
        if self.byte_pixel_threshold is not None:
            if not isinstance(self.byte_pixel_threshold, int):
                raise TypeError("byte_pixel_threshold must be an integer")
            if self.byte_pixel_threshold < 0:
                raise ValueError("byte_pixel_threshold must be non-negative")

        # Basic checks complete - make more checks and create directories where necessary
        self.input_paths = find_input_paths(self.input_path, "**/*.*", self.input_file_list)

    def to_table_properties(
        self,
        total_rows: int,
        highest_order: int,
        moc_sky_fraction: float,
        column_names=None,
    ) -> TableProperties:
        """Catalog-type-specific dataset info."""
        info = self.extra_property_dict() | {
            "catalog_name": self.output_artifact_name,
            "catalog_type": self.catalog_type,
            "total_rows": total_rows,
            "ra_column": self.ra_column,
            "dec_column": self.dec_column,
            "hats_cols_sort": self.sort_columns,
            "hats_max_rows": self.pixel_threshold,
            "hats_order": highest_order,
            "moc_sky_fraction": f"{moc_sky_fraction:0.5f}",
            "hats_npix_suffix": self.npix_suffix,
        }
        if self.should_write_skymap:
            info.update(
                {
                    "hats_skymap_order": self.mapping_healpix_order,
                    "hats_skymap_alt_orders": self.skymap_alt_orders,
                }
            )
        if self.add_healpix_29:
            info.update(
                {
                    "hats_col_healpix": SPATIAL_INDEX_COLUMN,
                    "hats_col_healpix_order": SPATIAL_INDEX_ORDER,
                }
            )
        properties = TableProperties(**info)

        if properties.default_columns and column_names:
            missing_columns = set(properties.default_columns) - set(column_names)
            if len(missing_columns):
                raise ValueError(f"Some default columns not found in catalog: {missing_columns}")

        return properties

    @classmethod
    def reimport_from_hats(cls, path: str | Path | UPath, output_dir: str | Path | UPath, **kwargs) -> Self:
        """Generate the import arguments to reimport a HATS catalog with different parameters

        Args:
            path (str | Path | UPath): the path to the existing HATS catalog to reimport
            output_dir (str | Path | UPath): the path to output the reimported catalog to
            kwargs: any import arguments to update from the existing catalog. Can be any argument usually
                passed to :func:`ImportArguments`

        Returns:
            A ImportArguments object with the arguments from the existing catalog, and any updates from kwargs
        """

        path = get_upath(path)

        catalog = hats.read_hats(path)

        if isinstance(catalog, CatalogCollection):
            path = catalog.main_catalog_dir
            catalog = catalog.main_catalog
        if not is_valid_catalog(path, strict=True):
            raise ValueError("path not a valid catalog")

        column_names = None
        if catalog.schema is not None:
            column_names = [name for name in catalog.schema.names if name not in HIVE_COLUMNS]

        in_file_paths = list(
            (path / DATASET_DIR).rglob(f"{PARTITION_ORDER}*/**/*{catalog.catalog_info.npix_suffix}")
        )

        addl_hats_properties = catalog.catalog_info.extra_dict(by_alias=True)

        addl_hats_properties.update(
            {
                "hats_cols_default": catalog.catalog_info.default_columns,
                "hats_npix_suffix": catalog.catalog_info.npix_suffix,
            }
        )

        if "addl_hats_properties" in kwargs:
            addl_hats_properties.update(kwargs.pop("addl_hats_properties"))

        existing_pixels = None
        if "existing_pixels" in kwargs:
            existing_pixels = kwargs.pop("existing_pixels")

        import_args = {
            "catalog_type": catalog.catalog_info.catalog_type,
            "ra_column": catalog.catalog_info.ra_column,
            "dec_column": catalog.catalog_info.dec_column,
            "input_file_list": in_file_paths,
            "file_reader": ParquetPyarrowReader(column_names=column_names),
            "output_artifact_name": catalog.catalog_name,
            "output_path": output_dir,
            "use_healpix_29": True,
            "add_healpix_29": False,
            "use_schema_file": hats.io.paths.get_common_metadata_pointer(catalog.catalog_base_dir),
            "expected_total_rows": catalog.catalog_info.total_rows,
            "addl_hats_properties": addl_hats_properties,
            "existing_pixels": existing_pixels,
        }

        import_args.update(**kwargs)
        return cls(**import_args)  # type: ignore



"""Import a set of non-hats files using dask for parallelization

Methods in this file set up a dask pipeline using futures.
The actual logic of the map reduce is in the `map_reduce.py` file.
"""

import os

import cloudpickle
import hats.io.file_io as io
import nested_pandas as npd
from hats.catalog import PartitionInfo
from hats.io import paths
from hats.io.parquet_metadata import write_parquet_metadata
from hats.io.skymap import write_skymap
from hats.io.validation import is_valid_catalog

import hats_import.catalog.map_reduce as mr
from hats_import.catalog.arguments import ImportArguments
from hats_import.catalog.resume_plan import ResumePlan


# pylint: disable=too-many-statements
def run(args, client):
    """Run catalog creation pipeline."""
    if not args:
        raise ValueError("args is required and should be type ImportArguments")
    if not isinstance(args, ImportArguments):
        raise ValueError("args must be type ImportArguments")

    resume_plan = ResumePlan(import_args=args)

    pickled_reader_file = os.path.join(resume_plan.tmp_path, "reader.pickle")
    with open(pickled_reader_file, "wb") as pickle_file:
        cloudpickle.dump(args.file_reader, pickle_file)

    if resume_plan.should_run_mapping:
        futures = []
        for key, file_path in resume_plan.map_files:
            futures.append(
                client.submit(
                    mr.map_to_pixels,
                    input_file=file_path,
                    resume_path=resume_plan.tmp_path,
                    pickled_reader_file=pickled_reader_file,
                    mapping_key=key,
                    highest_order=args.mapping_healpix_order,
                    ra_column=args.ra_column,
                    dec_column=args.dec_column,
                    use_healpix_29=args.use_healpix_29,
                    threshold_mode=resume_plan.threshold_mode,
                )
            )
        resume_plan.wait_for_mapping(futures)

    with resume_plan.print_progress(total=2, stage_name="Binning") as step_progress:
        # Check total rows matches expectation.
        raw_histogram_row_count = resume_plan.read_histogram(
            args.mapping_healpix_order, which_histogram="row_count"
        )
        total_rows = int(raw_histogram_row_count.sum())
        if args.expected_total_rows > 0 and args.expected_total_rows != total_rows:
            raise ValueError(
                f"Number of rows ({total_rows}) does not match expectation ({args.expected_total_rows})"
            )

        # Read in mem_size histogram if we will be thresholding by memory size.
        if resume_plan.threshold_mode == "mem_size":
            raw_histogram_mem_size = resume_plan.read_histogram(
                args.mapping_healpix_order, which_histogram="mem_size"
            )
        else:
            raw_histogram_mem_size = None

        # Get alignment file.
        step_progress.update(1)
        threshold = (
            args.pixel_threshold if resume_plan.threshold_mode == "row_count" else args.byte_pixel_threshold
        )
        alignment_file = resume_plan.get_alignment_file(
            raw_histogram_row_count,
            args.constant_healpix_order,
            args.highest_healpix_order,
            args.lowest_healpix_order,
            threshold,
            args.drop_empty_siblings,
            total_rows,
            args.existing_pixels,
            raw_histogram_mem_size,
        )

        step_progress.update(1)

    if resume_plan.should_run_splitting:
        futures = []
        for key, file_path in resume_plan.split_keys:
            futures.append(
                client.submit(
                    mr.split_pixels,
                    input_file=file_path,
                    pickled_reader_file=pickled_reader_file,
                    highest_order=args.mapping_healpix_order,
                    ra_column=args.ra_column,
                    dec_column=args.dec_column,
                    splitting_key=key,
                    cache_shard_path=args.tmp_path,
                    resume_path=resume_plan.tmp_path,
                    alignment_file=alignment_file,
                    use_healpix_29=args.use_healpix_29,
                )
            )

        resume_plan.wait_for_splitting(futures)

    if resume_plan.should_run_reducing:
        futures = []
        for (
            destination_pixel,
            source_pixel_count,
            destination_pixel_key,
        ) in resume_plan.get_reduce_items():
            futures.append(
                client.submit(
                    mr.reduce_pixel_shards,
                    cache_shard_path=args.tmp_path,
                    resume_path=resume_plan.tmp_path,
                    reducing_key=destination_pixel_key,
                    destination_pixel_order=destination_pixel.order,
                    destination_pixel_number=destination_pixel.pixel,
                    destination_pixel_size=source_pixel_count,
                    output_path=args.catalog_path,
                    ra_column=args.ra_column,
                    dec_column=args.dec_column,
                    sort_columns=args.sort_columns,
                    add_healpix_29=args.add_healpix_29,
                    use_schema_file=args.use_schema_file,
                    use_healpix_29=args.use_healpix_29,
                    delete_input_files=args.delete_intermediate_parquet_files,
                    write_table_kwargs=args.write_table_kwargs,
                    row_group_kwargs=args.row_group_kwargs,
                    npix_suffix=args.npix_suffix,
                    npix_parquet_name=args.npix_parquet_name,
                )
            )

        resume_plan.wait_for_reducing(futures)

    # All done - write out the metadata
    if resume_plan.should_run_finishing:
        num_steps = 6 if args.should_write_skymap else 5
        with resume_plan.print_progress(total=num_steps, stage_name="Finishing") as step_progress:
            partition_info = PartitionInfo.from_healpix(resume_plan.get_destination_pixels())
            partition_info_file = paths.get_partition_info_pointer(args.catalog_path)
            partition_info.write_to_file(partition_info_file)
            step_progress.update(1)

            column_names = []
            if not args.debug_stats_only:
                parquet_rows = write_parquet_metadata(
                    args.catalog_path,
                    create_thumbnail=args.create_thumbnail,
                    thumbnail_threshold=threshold,
                    create_metadata=args.create_metadata,
                )
                if total_rows > 0 and parquet_rows != total_rows:
                    raise ValueError(
                        f"Number of rows in parquet ({parquet_rows}) "
                        f"does not match expectation ({total_rows})"
                    )
                nested_schema = npd.read_parquet(paths.get_common_metadata_pointer(args.catalog_path))
                column_names = list(nested_schema.columns) + nested_schema.get_subcolumns()
            step_progress.update(1)
            if args.should_write_skymap:
                io.write_fits_image(
                    raw_histogram_row_count, paths.get_point_map_file_pointer(args.catalog_path)
                )
                write_skymap(
                    histogram=raw_histogram_row_count,
                    catalog_dir=args.catalog_path,
                    orders=args.skymap_alt_orders,
                )
                step_progress.update(1)
            catalog_info = args.to_table_properties(
                total_rows,
                partition_info.get_highest_order(),
                partition_info.calculate_fractional_coverage(),
                column_names=column_names,
            )
            catalog_info.to_properties_file(args.catalog_path)
            step_progress.update(1)
            resume_plan.clean_resume_files()
            step_progress.update(1)
            assert is_valid_catalog(args.catalog_path)
            step_progress.update(1)



from __future__ import annotations

from pathlib import Path

from hats.io.file_io import get_upath
from upath import UPath


def append_paths_to_pointer(pointer: str | Path | UPath, *paths: str) -> UPath:
    """Append directories and/or a file name to a specified file pointer.

    Parameters
    ----------
    pointer : str | Path | UPath
        `FilePointer` object to add path to
    *paths: str
        any number of directory names optionally followed by a file name to append to the
        pointer

    Returns
    -------
    UPath
        New file pointer to path given by joining given pointer and path names
    """
    pointer = get_upath(pointer)
    return pointer.joinpath(*paths)


def find_files_matching_path(pointer: str | Path | UPath, *paths: str) -> list[UPath]:
    """Find files or directories matching the provided path parts.

    Parameters
    ----------
    pointer : str | Path | UPath
        base File Pointer in which to find contents
    *paths: str
        any number of directory names optionally followed by a file name.
        directory or file names may be replaced with `*` as a matcher.

    Returns
    -------
    list[UPath]
        New file pointers to files found matching the path
    """
    pointer = get_upath(pointer)

    if len(paths) == 0:
        return [pointer]

    matcher = pointer.fs.sep.join(paths)
    contents = []
    for child in pointer.rglob(matcher):
        contents.append(child)

    if len(contents) == 0:
        return []

    contents.sort()
    return contents


def directory_has_contents(pointer: str | Path | UPath) -> bool:
    """Checks if a directory already has some contents (any files or subdirectories)

    Parameters
    ----------
    pointer : str | Path | UPath
        File Pointer to check for existing contents

    Returns
    -------
    bool
        True if there are any files or subdirectories below this directory.
    """
    pointer = get_upath(pointer)

    return next(pointer.rglob("*"), None) is not None



"""Convert a hipscatted catalog into a HATS catalog, with appropriate metadata/properties."""

import json
import tempfile
from typing import no_type_check

import hats.pixel_math.healpix_shim as healpix
import numpy as np
import pyarrow.parquet as pq
from dask.distributed import as_completed, get_worker
from dask.distributed import print as dask_print
from hats.catalog import CatalogType, PartitionInfo, TableProperties
from hats.io import file_io, parquet_metadata, paths, size_estimates

from hats_import._version import __version__ as build_version
from hats_import.hipscat_conversion.arguments import ConversionArguments
from hats_import.pipeline_resume_plan import print_progress


@no_type_check
def run(args: ConversionArguments, client):
    """Run index creation pipeline."""
    if not args:
        raise TypeError("args is required and should be type ConversionArguments")
    if not isinstance(args, ConversionArguments):
        raise TypeError("args must be type ConversionArguments")

    # Create basic properties, using catalog info, provenance info, and partition_info files
    catalog_info = None
    with (args.input_catalog_path / "catalog_info.json").open("r", encoding="utf-8") as json_file:
        catalog_info = json.load(json_file)
    provenance_info = None
    with (args.input_catalog_path / "provenance_info.json").open("r", encoding="utf-8") as json_file:
        provenance_info = json.load(json_file)

    catalog_type = CatalogType(catalog_info["catalog_type"])
    if catalog_type not in (
        CatalogType.OBJECT,
        CatalogType.SOURCE,
        CatalogType.ASSOCIATION,
    ):
        raise ValueError(
            "Conversion only implemented for object, source, and association tables. "
            "Other tables should be re-generated."
        )

    catalog_info.pop("epoch", None)
    catalog_info = catalog_info | args.extra_property_dict()
    if "tool_args" in provenance_info:
        builder_str = (
            provenance_info["tool_args"]["tool_name"]
            + " v"
            + provenance_info["tool_args"]["version"]
            + " hats-importer conversion v"
            + build_version
        )
        catalog_info["hats_builder"] = builder_str
        if runtime_args := provenance_info["tool_args"].get("runtime_args"):
            catalog_info["hats_cols_sort"] = runtime_args.get("sort_columns")
            catalog_info["hats_cols_survey_id"] = runtime_args.get("sort_columns")
            catalog_info["hats_max_rows"] = runtime_args.get("pixel_threshold")

    partition_info = PartitionInfo.read_from_dir(args.input_catalog_path)
    catalog_info["hats_order"] = partition_info.get_highest_order()

    properties = TableProperties(**catalog_info)

    schema = file_io.read_parquet_metadata(
        args.input_catalog_path / "_common_metadata"
    ).schema.to_arrow_schema()

    futures = []
    for pixel in partition_info.get_healpix_pixels():
        futures.append(
            client.submit(
                _convert_partition_file, pixel, args, schema, properties.ra_column, properties.dec_column
            )
        )
    for future in print_progress(
        as_completed(futures),
        stage_name="Converting Parquet",
        total=len(futures),
        use_progress_bar=args.progress_bar,
        simple_progress_bar=args.simple_progress_bar,
    ):
        if future.status == "error":
            raise future.exception()

    with print_progress(
        total=4,
        stage_name="Finishing",
        use_progress_bar=args.progress_bar,
        simple_progress_bar=args.simple_progress_bar,
    ) as step_progress:
        total_rows = parquet_metadata.write_parquet_metadata(
            args.catalog_path,
            create_thumbnail=(catalog_type in (CatalogType.OBJECT, CatalogType.SOURCE)),
            thumbnail_threshold=int(getattr(properties, "hats_max_rows", 1_000_000)),
        )
        if total_rows != properties.total_rows:
            raise ValueError(
                f"Unexpected number of rows (original: {properties.total_rows}"
                f" written to parquet: {total_rows})"
            )
        step_progress.update(1)
        file_io.remove_directory(args.tmp_path, ignore_errors=True)
        step_progress.update(1)
        ## Update total size with newly-written parquet files.
        properties.__pydantic_extra__["hats_estsize"] = size_estimates.estimate_dir_size(
            args.catalog_path, divisor=1024
        )
        properties.to_properties_file(args.catalog_path)
        partition_info.write_to_file(args.catalog_path / "partition_info.csv")
        step_progress.update(1)
        _write_nested_fits_map(args.input_catalog_path, args.catalog_path)
        step_progress.update(1)


def _convert_partition_file(pixel, args, schema, ra_column, dec_column):
    try:
        # Paths are changed between hipscat and HATS!
        input_file = (
            args.input_catalog_path
            / f"Norder={pixel.order}"
            / f"Dir={pixel.dir}"
            / f"Npix={pixel.pixel}.parquet"
        )

        table = pq.read_table(input_file.path, filesystem=input_file.fs, schema=schema)

        table = table.drop_columns(["_hipscat_index", "Norder", "Dir", "Npix"]).add_column(
            0,
            "_healpix_29",
            [
                healpix.radec2pix(
                    29,
                    table[ra_column].to_numpy(),
                    table[dec_column].to_numpy(),
                )
            ],
        )
        table = table.replace_schema_metadata()

        destination_file = paths.pixel_catalog_file(args.catalog_path, pixel)
        destination_file.parent.mkdir(parents=True, exist_ok=True)
        pq.write_table(table, destination_file.path, filesystem=destination_file.fs)
    except Exception as exception:  # pylint: disable=broad-exception-caught
        try:
            dask_print("  worker address:", get_worker().address)
        except Exception:  # pylint: disable=broad-exception-caught
            pass
        dask_print(exception)
        raise exception


# pylint: disable=import-outside-toplevel
def _write_nested_fits_map(input_dir, output_dir):
    # Healpy is an optional dependency, used only for reads of legacy fits files.
    import healpy as hp  # pylint: disable=import-error

    input_file = input_dir / "point_map.fits"
    if not input_file.exists():
        return
    with tempfile.NamedTemporaryFile() as _tmp_file:
        with input_file.open("rb") as _map_file:
            map_data = _map_file.read()
            _tmp_file.write(map_data)
            map_fits_image = hp.read_map(_tmp_file.name, nest=True, h=True)
            header_dict = dict(map_fits_image[1])
            if header_dict["ORDERING"] != "NESTED":
                map_fits_image = hp.read_map(_tmp_file.name)
            else:
                map_fits_image = map_fits_image[0]
    map_fits_image = map_fits_image.astype(np.int32)

    file_io.write_fits_image(map_fits_image, output_dir / "point_map.fits")



import pyarrow.parquet as pq
import pytest
from hats import read_hats

from hats_import.collection.arguments import CollectionArguments
from hats_import.collection.run_import import run


def test_empty_args():
    """Runner should fail with empty arguments"""
    with pytest.raises(TypeError, match="CollectionArguments"):
        run(None, None)


def test_bad_args():
    """Runner should fail with mis-typed arguments"""
    args = {"output_artifact_name": "bad_arg_type"}
    with pytest.raises(TypeError, match="CollectionArguments"):
        run(args, None)


@pytest.mark.dask(timeout=150)
def test_import_collection(
    dask_client,
    small_sky_source_dir,
    tmp_path,
):
    args = (
        CollectionArguments(
            output_artifact_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            input_path=small_sky_source_dir,
            file_reader="csv",
            catalog_type="source",
            ra_column="source_ra",
            dec_column="source_dec",
            sort_columns="source_id",
            highest_healpix_order=2,
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=50.0)
        .add_index(indexing_column="object_id", include_healpix_29=False)
        .add_index(indexing_column="source_id", include_healpix_29=False)
    )
    run(args, dask_client)

    collection = read_hats(tmp_path / "small_sky")
    assert collection.collection_path == args.catalog_path

    assert collection.all_margins == ["small_sky_5arcs", "small_sky_50arcs"]
    assert collection.all_indexes == {"object_id": "small_sky_object_id", "source_id": "small_sky_source_id"}

    catalog = read_hats(tmp_path / "small_sky" / "small_sky")
    assert catalog.on_disk
    assert catalog.catalog_info.ra_column == "source_ra"
    assert catalog.catalog_info.dec_column == "source_dec"

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_5arcs")
    assert catalog.on_disk
    assert len(catalog.get_healpix_pixels()) == 2
    assert len(catalog) == 4

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_50arcs")
    assert catalog.on_disk
    assert len(catalog.get_healpix_pixels()) == 2
    assert len(catalog) == 17

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_object_id")
    ## This isn't 131, because one object's sources spans two pixels.
    assert catalog.catalog_info.total_rows == 132

    catalog = read_hats(tmp_path / "small_sky" / "small_sky_source_id")
    assert catalog.catalog_info.total_rows == 17161

    ## Re-running shouldn't create any new tables.
    args = (
        CollectionArguments(
            output_artifact_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            input_path=small_sky_source_dir,
            file_reader="csv",
            catalog_type="source",
            ra_column="source_ra",
            dec_column="source_dec",
            sort_columns="source_id",
            highest_healpix_order=2,
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=50.0)
        .add_index(indexing_column="object_id", include_healpix_29=False)
        .add_index(indexing_column="source_id", include_healpix_29=False)
    )
    run(args, dask_client)

    collection = read_hats(tmp_path / "small_sky")
    assert collection.collection_path == args.catalog_path

    ## Don't worry about order for this equality test.
    assert set(collection.all_margins) == set(["small_sky_50arcs", "small_sky_5arcs"])
    assert collection.all_indexes == {"object_id": "small_sky_object_id", "source_id": "small_sky_source_id"}


@pytest.mark.dask(timeout=150)
def test_import_collection_resume_supplemental(
    dask_client,
    small_sky_source_dir,
    tmp_path,
):
    args = (
        CollectionArguments(
            output_artifact_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            input_path=small_sky_source_dir,
            file_reader="csv",
            catalog_type="source",
            ra_column="source_ra",
            dec_column="source_dec",
            sort_columns="source_id",
            highest_healpix_order=2,
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=50.0)
        .add_index(indexing_column="object_id", include_healpix_29=False)
        .add_index(indexing_column="not_a_good_column", include_healpix_29=False)
    )
    with pytest.raises(ValueError, match="not_a_good_column"):
        run(args, dask_client)

    ## Re-running should complete successfully.
    args = (
        CollectionArguments(
            output_artifact_name="small_sky",
            output_path=tmp_path,
            progress_bar=False,
        )
        .catalog(
            input_path=small_sky_source_dir,
            file_reader="csv",
            catalog_type="source",
            ra_column="source_ra",
            dec_column="source_dec",
            sort_columns="source_id",
            highest_healpix_order=2,
        )
        .add_margin(margin_threshold=5.0)
        .add_margin(margin_threshold=50.0)
        .add_index(indexing_column="object_id", include_healpix_29=False)
        .add_index(indexing_column="source_id", include_healpix_29=False)
    )
    ## Margins and one index have already been created successfully,
    ## and should not be modified by this execution.
    assert len(args.get_margin_args()) == 0
    assert len(args.get_index_args()) == 1
    run(args, dask_client)

    collection = read_hats(tmp_path / "small_sky")
    assert collection.collection_path == args.catalog_path

    ## Don't worry about order for this equality test.
    ## What we really are testing here is that the margins are included in the
    ## Collection properties, even though they were created by a previous (unsuccessful) run
    ## of the pipeline, and were not included in the properties file at that time.
    assert set(collection.all_margins) == set(["small_sky_50arcs", "small_sky_5arcs"])
    assert collection.all_indexes == {"object_id": "small_sky_object_id", "source_id": "small_sky_source_id"}


@pytest.mark.dask(timeout=150)
def test_import_collection_healpix13(
    dask_client,
    formats_dir,
    tmp_path,
):
    args = (
        CollectionArguments(
            output_artifact_name="small_sky_healpix13",
            output_path=tmp_path,
            tmp_dir=tmp_path,
            # Also test that we plumb npix parameters through collection arguments.
            npix_suffix="/",
            npix_parquet_name="today.parquet",
        )
        .catalog(
            input_file_list=formats_dir / "small_sky_healpix13.csv",
            file_reader="csv",
            pixel_threshold=3000,
            row_group_kwargs={"num_rows": 1_000},
            highest_healpix_order=2,
            add_healpix_29=False,
            addl_hats_properties={"hats_col_healpix": "healpix13", "hats_col_healpix_order": 13},
            constant_healpix_order=1,
        )
        .add_margin(margin_threshold=3600, is_default=True)
        .add_margin(margin_threshold=7200)
        .add_index(
            indexing_column="id",
            include_healpix_29=False,
            compute_partition_size=200_000,
        )
    )

    run(args, dask_client)

    collection = read_hats(tmp_path / "small_sky_healpix13")

    catalog = collection.main_catalog
    assert catalog.catalog_info.healpix_column == "healpix13"
    assert catalog.catalog_info.healpix_order == 13
    assert catalog.has_healpix_column()

    test_file = (
        tmp_path
        / "small_sky_healpix13"
        / "small_sky_healpix13"
        / "dataset"
        / "Norder=1"
        / "Dir=0"
        / "Npix=44"
        / "today.parquet"
    )
    assert test_file.exists()
    metadata = pq.read_metadata(test_file)
    assert metadata.num_row_groups == 1
    assert metadata.row_group(0).column(0).compression == "ZSTD"

    catalog = read_hats(tmp_path / "small_sky_healpix13" / collection.default_margin)
    assert catalog.catalog_info.healpix_column == "healpix13"
    assert catalog.catalog_info.healpix_order == 13
    assert catalog.has_healpix_column()

    test_file = (
        tmp_path
        / "small_sky_healpix13"
        / "small_sky_healpix13_1deg"
        / "dataset"
        / "Norder=1"
        / "Dir=0"
        / "Npix=44"
        / "today.parquet"
    )
    assert test_file.exists()
    metadata = pq.read_metadata(test_file)
    assert metadata.num_row_groups == 1
    assert metadata.row_group(0).column(0).compression == "ZSTD"
